#include <iostream>

#include <math.h>

#include "support/binarysearch.h"
#include "imagesource/imagesource_modifiedgamma.h"

#include "config.h" // For HAVE_LIBPTHREADS
#include "argyll_support.h"
#include "channelresponse.h"
#include "1dinterp.h"

using namespace std;


// Class used to search for hue deviation and set ink limit.
// The algorith used was suggested by Graeme Gill - we measure
// the deviation by considerint two points on the curve.
// We find the distance from the paper colour to each point, subtract
// one from the other, and divide by the distance between the points.
// The result tends towards 1.0 while the hue is constant, and drops as
// paper saturation causes the hue to wander.

class BinarySearch_HueSlope : public BinarySearch
{
	public:
	BinarySearch_HueSlope(Interpolate_Lab &lab) : BinarySearch(), labint(lab)
	{
		origin=labint[0.0];
	}
	virtual ~BinarySearch_HueSlope()
	{
	}
	virtual double Evaluate(double input)
	{
		LabValue test=labint[0.0];
		double pi=input-1.0;
		double ci=input+1.0;
		LabValue current=labint[ci];
		LabValue prev=labint[pi];

		double d1=prev.dE(origin);
		double d2=current.dE(origin);
		double d3=current.dE(prev);
		double slope=(d2-d1)/d3;
		return(slope);
	}
	protected:
	Interpolate_Lab &labint;
	LabValue origin;
};


// Class to search for a given dL / d_input.  The change in L* for each
// increment in ink coverage will reduce as the paper starts to saturate
// so we search for the point where this dL / d_input falls below a certain
// threshold, and use this as the ink limit for black, since we can't use hue
// deviation for black!
// The threshold is derived from the colour channels.

class BinarySearch_dLbydi : public BinarySearch
{
	public:
	BinarySearch_dLbydi(Interpolate_Lab &lab,double gamma) : BinarySearch(), labint(lab), gamma(gamma)
	{
		origin=labint[0.0];
	}
	virtual ~BinarySearch_dLbydi()
	{
	}
	virtual double Evaluate(double input)
	{
		double pi=input-1.0;
		double ci=input+1.0;
		LabValue current=labint[ci];
		LabValue prev=labint[pi];

		double d3=prev.L-current.L;
		double di=ForwardGamma(ci/100.0)-ForwardGamma(pi/100.0);
		double dEbydi=d3/di;
		return(dEbydi);
	}
	double ForwardGamma(double in)
	{
		return(1.0-ImageSource_ModifiedGamma::ModifiedGamma(1.0-in,gamma));
	}
	protected:
	Interpolate_Lab &labint;
	double gamma;
	LabValue origin;
};


////////////////////////////////////////////////////////////////////////////////////


RawChannelResponses::RawChannelResponses(const char *filename,double gamma,double smoothing)
	: TI3File(filename), firstchannel(NULL), paper(), gamma(gamma)
{
	// Deduce paper-white from the data.  This only works for Subtractive colour,
	// since we check for colorants all set to zero.  Since there are commonly several
	// unprinted patches, we average them.

	int papercount=0;
	for(int i=0;i<GetSets();++i)
	{
		double acc=0.0;
		double **data=GetData(i);
		for(int j=0;j<colorants;++j)
		{
			acc+=*data[j];
		}
		if(acc==0.0)
		{
			paper.X+=*data[Xindex];
			paper.Y+=*data[Yindex];
			paper.Z+=*data[Zindex];
			++papercount;
		}
	}
	if(papercount)
	{
		paper.X/=papercount;
		paper.Y/=papercount;
		paper.Z/=papercount;
	}
	cerr << papercount << " unprinted patches found - Paper colour (XYZ): " << paper.X << ", " << paper.Y << ", " << paper.Z << endl;

	// Generate the channel responses - one for each colorant
	for(int i=0;i<colorants;++i)
	{
		new ChannelResponse(this,i,smoothing);
	}

	CalculateInkLimits();
}


RawChannelResponses::~RawChannelResponses()
{
	while(firstchannel)
		delete firstchannel;
}


ChannelResponse *RawChannelResponses::FirstChannel()
{
	return(firstchannel);
}


ChannelResponse *RawChannelResponses::GetChannel(int c)
{
	ChannelResponse *result=FirstChannel();
	while(result)
	{
		if(result->channelindex==c)
			return(result);
		result=result->NextChannel();
	}
	return(result);
}


ChannelResponse *RawChannelResponses::GetChannel(const char *name)
{
	int idx=GetColorantIndex(name);
	return(GetChannel(idx));
}


ChannelResponseSample *RawChannelResponses::GetPaper()
{
	return(&paper);
}


void RawChannelResponses::CalculateInkLimits()
{
	// First loop through the colour channels, determining limits using
	// a/b angle deviation...
	ChannelResponse *c=FirstChannel();
	while(c)
	{
		if(c->GetType()!=CHANNELRESPONSETYPE_BLACK)
			c->CalculateInkLimit();
		c=c->NextChannel();
	}

	// Then loop through black channels.  We'll use the limits of the colour
	// channels already processed to determine a suitable dE/dinput threshold
	// for the black channels.
	c=FirstChannel();
	while(c)
	{
		if(c->GetType()==CHANNELRESPONSETYPE_BLACK)
			c->CalculateInkLimit();
		c=c->NextChannel();
	}
}


////////////////////////////////////////////////////////////////////////////////////////////////


ChannelResponse::ChannelResponse(RawChannelResponses *header,int channel,double smoothing)
	: ti3index(channel), channelname(NULL), samplecount(0), samples(NULL),
	header(header), nextchannel(NULL), prevchannel(NULL), labinterp(), labinterp_smoothed(),
	type(CHANNELRESPONSETYPE_OTHER)
{
	// We sort out the colorant stuff first because if an unknown colorant is going
	// to trigger an exception in the DeviceNColorant constructor we want it to
	// happen before we're linked into the list.
	channelname=header->GetFieldName(channel);
	cerr << "Creating channel for: " << channelname << endl;
	const char *sn=GetShortName();
	const char *longname=ArgyllSupport::GetColorantLongName(sn);

	// Note: channelindex and channel are not neccesarily the same.
	channelindex=header->GetColorantIndex(longname);

	cerr << "Creating channel: " << longname << ", channel " << channelindex << endl;

	// Link this node into the list

	if((prevchannel=header->firstchannel))
	{
		while(prevchannel->nextchannel)
			prevchannel=prevchannel->nextchannel;
		prevchannel->nextchannel=this;
	}
	else
		header->firstchannel=this;


	// We obtain the raw data from the .ti3 file and count the number
	// of patches which involve just the current channel

	for(int i=0;i<header->GetSets();++i)
	{
		double chan=0.0;
		double acc=0.0;
		double **data=header->GetData(i);
		for(int j=0;j<header->colorants;++j)
		{
			if(j==ti3index)
				chan=*data[j];
			else
				acc+=*data[j];
		}
		// Chan contains the contents of the row for this channel
		// acc contains the sum total of the other colorants' values.
		if(acc<0.01 && chan>0.01)

		{
			++samplecount;
		}
	}

	if(samplecount<3)	// No point trying to work with fewer than 3 samples.
	{
		// Need to mark the channel as inactive in some way.  Merely deleting it works
		// as long as the inactive channel is the last one - but will cause problems if
		// channels are referenced by index.

		delete this;	// FIXME - double-check legality of doing this in a constructor
		return;
	}

	++samplecount;	// The count calculated above excludes the paper colour


	// Now we allocate storage for the patches we identified;
	samples=new ChannelResponseSample[samplecount];
	int idx=0;
	samples[0]=header->paper;
	++idx;

	// Set min and max values for a and b using the paper colour.
	// Probably no need for this - we can probably assume paper has max L / min b
	// and the last sample has min L / max b.
	LabValue l(samples[0]);
	float lmin=l.L;
	float lmax=l.L;
	float bmin=l.b;
	float bmax=l.b;


	// Loop through the raw data again, storing the samples this time.

	for(int i=0;i<header->GetSets();++i)
	{
		double chan=0.0;
		double acc=0.0;
		double **data=header->GetData(i);
		for(int j=0;j<header->colorants;++j)
		{
			if(j==ti3index)
				chan=*data[j];
			else
				acc+=*data[j];
		}
		// Chan contains the contents of the row for this channel
		// acc contains the sum total of the other colorants' values.
		if(acc<0.01 && chan>0.01)
		{
			samples[idx].input=chan;
			samples[idx].X=*data[header->Xindex];
			samples[idx].Y=*data[header->Yindex];
			samples[idx].Z=*data[header->Zindex];

			LabValue l(samples[idx]);
			if(l.L < lmin) lmin=l.L;
			if(l.L > lmax) lmax=l.L;
			if(l.b < bmin) bmin=l.b;
			if(l.b > bmax) bmax=l.b;

			++idx;
		}
	}

	labinterp.Populate(samplecount,samples);
	labinterp_smoothed.Populate(samplecount,samples,smoothing);	// Apply smoothing

	cerr << "L* range: " << lmin << " - " << lmax << endl;
	cerr << "b range: " << bmin << " - " << bmax << endl;
	if((bmax-bmin)/3.0 > (lmax-lmin))
	{
		// If b range is larger than L range, we probably have a yellow
		// channel.  It's arguably better to linearize against b than L*
		// in this case.
		cerr << "b has greater range than L* - probably Yellow channel" << endl;
		type=CHANNELRESPONSETYPE_YELLOW;
	}
	else
	{
		// Not a yellow channel - but is it a black channel?
		// Check that by comparing Lch c against a threshold.
		// If it's lower than the threshold, chances are it's a black channel.
		LabValue max(samples[samplecount-1]);
		double Lch_c=sqrt(max.a*max.a+max.b*max.b);
		if(Lch_c<20.0)
		{
			cerr << "Probably a black channel" << endl;
			type=CHANNELRESPONSETYPE_BLACK;
		}
	}

	// Initialize the ink limit - a more appropriate value will be either set by the client
	// or calculated by the ChannelResponses header
	SetInkLimit(100.0);

//	// Create DeviceNColorant which provides RGB equivalent colours for preview rendering.
//	new DeviceNColorant(*header,longname);
}


ChannelResponse::~ChannelResponse()
{
	if(samples)
		delete[] samples;

	if(prevchannel)
		prevchannel->nextchannel=nextchannel;
	else
		header->firstchannel=nextchannel;
	if(nextchannel)
		nextchannel->prevchannel=prevchannel;
}


enum ChannelResponseType ChannelResponse::GetType()
{
	return(type);
}


void ChannelResponse::CalculateInkLimit()
{
	if(type==CHANNELRESPONSETYPE_BLACK)
	{
		// Cycle through the colours here, finding dL/di
		// at the ink limit, scaled by the channel's L* range.
		// We take the smallest of these, multiply by the black
		// channel's L* range and use it to determine the ink limit.
		// The colour channels have already been ink limited at this point.
		cerr << "Finding ink limit for black channel" << endl;
		double dLbydi_target=100.0;
		ChannelResponse *c=header->FirstChannel();
		while(c)
		{
			// We only want to consider channels which aren't either Yellow or Black.
			if(c->GetType()==CHANNELRESPONSETYPE_OTHER)
			{
				Interpolate_Lab *li=c->GetLabInterp();
				double lrange=(*li)[0.0].L-(*li)[100.0].L;
				BinarySearch_dLbydi search(*li,header->gamma);
				double dLbydi=search.Evaluate(c->GetInkLimit())/lrange;
				if(dLbydi<dLbydi_target)
					dLbydi_target=dLbydi;
			}
			c=c->NextChannel();
		}
		cerr << "Best target: " << dLbydi_target << endl;
		if(dLbydi_target==100.0)
			SetInkLimit(100);
		else
		{
			BinarySearch_dLbydi search(labinterp_smoothed,header->gamma);
			double lrange=labinterp[0.0].L-labinterp[100.0].L;
			double result=search.Search(1.0,99.0,lrange*dLbydi_target);
			cerr << "Suggested ink limit: " << result << endl;
			SetInkLimit(result);
		}
	}
	else
	{
		BinarySearch_HueSlope hueslope(labinterp_smoothed);
		double result=hueslope.Search(1.0,99.0,0.975);
		SetInkLimit(result);
		cerr << "Suggested ink limit: " << result << endl;
	}
}


ChannelResponse *ChannelResponse::NextChannel()
{
	return(nextchannel);
}


ChannelResponse *ChannelResponse::PrevChannel()
{
	return(prevchannel);
}


ChannelResponse *ChannelResponse::FindSubChannel(ChannelResponse *subchan)
{
	if(subchan)
		subchan=subchan->NextChannel();
	else
		subchan=header->FirstChannel();

	while(subchan)
	{
//		cerr << "Comparing " << GetLongName() << " against " << subchan->GetLongName() << endl;
		if(GetType()==subchan->GetType())
		{
			Interpolate_Lab *li2=subchan->GetLabInterp();
			LabValue lsub=(*li2)[50];	// Lab value of subchannel at 50% input

			if(lsub.L>labinterp[50].L)	// Subchannel must be lighter at 50% than main channel at 50% - otherwise mismatch or reversal.
			{
				double l=labinterp.ReverseInterpolate_L(lsub.L);
				LabValue lchan=labinterp[l];	// Lab value of main channel at closest point to subchannel at 50%.

				double de=lsub.dE(lchan);	// Distance between main and subchannel at 50% of subchannel.
				cerr << "dE: " << de << endl;
				if(de<35.0)
				{
					// We allow a much wider dE between black channels
					if(GetType()==CHANNELRESPONSETYPE_BLACK)
						return(subchan);

#if 0
					// Now to make sure the channels really are the same, we divide distance from origin by distance between points.
					cerr << "Checking hue angles: " << endl;

					LabValue p1=lsub-header->paper;		// Vector from paper to subchannel
					LabValue p2=lchan-header->paper;	// Vector from paper to main channel

//					double m1=sqrt(p1.a*p1.a+p1.b*p1.b);
//					p1.a/=m1; p1.b/=m1;					// Normalize and b;

//					double m2=sqrt(p2.a*p2.a+p2.b*p2.b);
//					p2.a/=m2; p2.b/=m2;					// Normalize and b;

					cerr << "P1: " << p1 << ", P2: " << p2 << endl;

					double d_a=p1.a-p2.a;
					double d_b=p1.b-p2.b;
					double d_ab=sqrt(d_a*d_a+d_b*d_b);

					cerr << "d_ab: " << d_ab << endl;
	
					double d_origin=lsub.dE(header->paper)-lchan.dE(header->paper);
					cerr << "Difference in distance from origin: " << d_origin << ", d_o / dE: " << d_origin/de << endl;
					if((d_origin/de)<2.5)
#endif
					if(de<10.0)
						return(subchan);
				}
			}
		}
		subchan=subchan->NextChannel();
	}
	return(NULL);
}


const char *ChannelResponse::GetName()
{
	return(channelname);
}


const char *ChannelResponse::GetLongName()
{
	return(ArgyllSupport::GetColorantLongName(GetShortName()));
}


const char *ChannelResponse::GetShortName()
{
	if(channelname)
	{
		int i=0;
		while(channelname[i])
		{
			if(channelname[i]=='_')
			{
				return(channelname+i+1);
			}
			++i;
		}
	}
	return(NULL);
}


int ChannelResponse::GetChannelIndex()
{
	return(channelindex);
}


int ChannelResponse::GetSampleCount()
{
	return(samplecount);
}


Interpolate_Lab *ChannelResponse::GetLabInterp()
{
	return(&labinterp);
}


Interpolate_Lab *ChannelResponse::GetLabInterp_Smoothed()
{
	return(&labinterp_smoothed);
}


double ChannelResponse::GetInkLimit()
{
	return(inklimit);
}


void ChannelResponse::SetInkLimit(double limit)
{
	inklimit=limit;
	min=labinterp[0.0];
	max=labinterp[limit];
	s_min=labinterp_smoothed[0.0];
	s_max=labinterp_smoothed[limit];
}


// Interpolate - input range: 0.0 - inklimit;
// Output range: 0.0 - 100.0
double ChannelResponse::Interpolate(double in)
{
	if(type==CHANNELRESPONSETYPE_YELLOW)
	{
		double r=max.b-min.b;
		LabValue l=labinterp[in];
		return(100*(l.b-min.b)/r);
	}
	else
	{
		double r=min.L-max.L;
		LabValue l=labinterp[in];
		return(100*(min.L-l.L)/r);
	}
}


double ChannelResponse::Interpolate_Smoothed(double in)
{
	if(type==CHANNELRESPONSETYPE_YELLOW)
	{
		double r=s_max.b-s_min.b;
		LabValue l=labinterp_smoothed[in];
		return(100*(l.b-s_min.b)/r);
	}
	else
	{
		double r=s_min.L-s_max.L;
		LabValue l=labinterp_smoothed[in];
		return(100*(s_min.L-l.L)/r);
	}
}


// ReverseInterpolate - input range: 0.0 - 100.0;
// Output range: 0.0 - inklimit
double ChannelResponse::ReverseInterpolate(double in)
{
	if(type==CHANNELRESPONSETYPE_YELLOW)
	{
		double r=max.b-min.b;
		in=(r*in)/100.0+min.b;
		double out=labinterp.ReverseInterpolate_b(in);
		return(out);
	}
	else
	{
		double r=min.L-max.L;
		in=min.L-(r*in)/100.0;
		double out=labinterp.ReverseInterpolate_L(in);
		return(out);
	}
}


// ReverseInterpolateSmoothed - input range: 0.0 - 100.0;
// Output range: 0.0 - inklimit
double ChannelResponse::ReverseInterpolate_Smoothed(double in)
{
	if(type==CHANNELRESPONSETYPE_YELLOW)
	{
		double r=s_max.b-s_min.b;
		in=(r*in)/100.0+s_min.b;
		double out=labinterp_smoothed.ReverseInterpolate_b(in);
		return(out);
	}
	else
	{
		double r=s_min.L-s_max.L;
		in=s_min.L-(r*in)/100.0;
		double out=labinterp_smoothed.ReverseInterpolate_L(in);
		return(out);
	}
}


ChannelResponseSample *ChannelResponse::GetSamples()
{
	return(samples);
}


ChannelResponseSample &ChannelResponse::operator[](int i)
{
	return(samples[i]);
}

