#ifndef CHANNELRESPONSE_H
#define CHANNELRESPONSE_H

#include "imagesource/devicencolorant.h"
#include "profilemanager/ciesupport.h"
#include "ticgats.h"
#include "labinterp.h"

class ChannelResponseSample : public XYZValue
{
	public:
	ChannelResponseSample() : XYZValue(), input(0.0) {}
	ChannelResponseSample(float input,float X,float Y, float Z) : XYZValue(X,Y,Z), input(input) {}
	~ChannelResponseSample() {}
	float input;
};


// Header class - creates a list of ChannelResponses, given the filename of a .ti3 file, and
// optionally, the gamma correction which was applied when printing the chart from which the
// .ti3 file was read.  This is needed because ink limits are calculated based on diminishing
// returns from applying more ink, so the raw uncorrected ink coverage is needed.

// We include a DeviceNColorantList for the RGB equivalent colours used for preview rendering.

class ChannelResponse;
class RawChannelResponses : public TI3File
{
	public:
	RawChannelResponses(const char *filename,double gamma=1.0,double smoothing=0.25);
	virtual ~RawChannelResponses();
	ChannelResponse *FirstChannel();
	ChannelResponse *GetChannel(int index);
	ChannelResponse *GetChannel(const char *name);
	ChannelResponseSample *GetPaper();
	void CalculateInkLimits();
	protected:
	ChannelResponse *firstchannel;
	ChannelResponseSample paper;
	double gamma;
	friend class ChannelResponse;
};


// We need to treat different types of channel differently.  Yellow is linearized by b rather than L*,
// and we can't use hue deviation to select an ink limit for black channels.
enum ChannelResponseType {CHANNELRESPONSETYPE_OTHER,CHANNELRESPONSETYPE_YELLOW,CHANNELRESPONSETYPE_BLACK};

// ChanneResponse - maintains data points for an individual channel, and builds interpolators with
// minimal and moderate smoothing.
// Also calculates suitable ink limit, and can provide forward and reverse interpolation
// using that ink limit.

class ChannelResponse
{
	public:
	ChannelResponse(RawChannelResponses *header,int channel,double smoothing=0.25);
	~ChannelResponse();
	ChannelResponse *NextChannel();
	ChannelResponse *PrevChannel();
	ChannelResponse *FindSubChannel(ChannelResponse *prev=NULL);
	const char *GetName();
	const char *GetShortName();
	const char *GetLongName();
	int GetChannelIndex();
	enum ChannelResponseType GetType();

										// Interpolate: input range 0.0 - ink limit
										// output range: 0.0 - 100.0.
	double Interpolate(double in);
	double Interpolate_Smoothed(double in);
										// ReverseInterpolate: input range 0.0 - 100.0
										// output range: 0.0 - ink limit 
	double ReverseInterpolate(double in);
	double ReverseInterpolate_Smoothed(double in);

	int GetSampleCount();				// Functions to obtain the raw channel samples
	ChannelResponseSample *GetSamples(); 
	ChannelResponseSample &operator[](int i);

	Interpolate_Lab *GetLabInterp();	// The LabInterp is based on the raw channel
										// data, and isn't modified by the ink limit.

	Interpolate_Lab *GetLabInterp_Smoothed();	// This secondary LabInterp is smoothed, which
												// improves the behaviour of the hue-deviation function
												// used for ink limiting.

	double GetInkLimit();
	void SetInkLimit(double limit);		// Adjusts ranges returned by Interpolate()
										// and ReverseInterpolate() functions.
										// limit takes the range 0.0 - 100.0
	void CalculateInkLimit();			// Calculates an ink limit based on a hue-deviation
										// threshold function, as suggested by Graeme Gill.
	protected:
	int channelindex;					// This channel's index into the DeviceNColorantList
	int ti3index;						// THis channel's index in the TI3 file
	const char *channelname;
	int samplecount;
	ChannelResponseSample *samples;
	RawChannelResponses *header;
	ChannelResponse *nextchannel,*prevchannel;
	Interpolate_Lab labinterp;
	Interpolate_Lab labinterp_smoothed;
	enum ChannelResponseType type;
	double inklimit;
	LabValue min,max;
	LabValue s_min,s_max; // min/max for the smoothed version of the data
	friend class RawChannelResponses;
};

#endif

