#include <iostream>

#include <gtk/gtk.h>

#include "miscwidgets/simplecombo.h"
#include "miscwidgets/patheditor.h"

#include "stpui_widgets/stpui_printerselector.h"
#include "stpui_widgets/stpui_optionbook.h"
#include "gp_cppsupport/printoutputselector.h"

#include "dialogs.h"
#include "lincore.h"
#include "argyll_support.h"
#include "argyll_uisupport.h"

#include "gettext.h"
#define _(x) gettext(x)


using namespace std;

#define RESPONSE_SAVE 1


struct prefsdialogdata
{
	LinearizationCore *core;
	GtkWidget *argyllpath;
	GtkWidget *targetpath;
	GtkWidget *instrument;
	GtkWidget *papersize;
	GtkWidget *resolution;
};


void PreferencesDialog(LinearizationCore *core,GtkWindow *parent)
{
	GtkWidget *dialog;
	GtkWidget *notebook;
	GtkWidget *vbox;
	GtkWidget *label;

	struct prefsdialogdata dd;
	dd.core=core;

	ArgyllSupport argyll(core->FindString("ArgyllPath"));

 	dialog=gtk_dialog_new_with_buttons(_("Preferences..."),
		parent,GtkDialogFlags(0),
		GTK_STOCK_SAVE,RESPONSE_SAVE,
		GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);

	gtk_window_set_default_size (GTK_WINDOW (dialog), 450,350);

	notebook=gtk_notebook_new();
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox),notebook,TRUE,TRUE,5);
	gtk_widget_show(GTK_WIDGET(notebook));


	// Path to Argyll commands

	label=gtk_label_new(_("Path to Argyll utilities"));
	gtk_widget_show(label);
	dd.argyllpath = patheditor_new(&argyll);
	gtk_widget_show(dd.argyllpath);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook),GTK_WIDGET(dd.argyllpath),GTK_WIDGET(label));


	// Path to Targets

	label=gtk_label_new(_("Path to GPLin Targets"));
	gtk_widget_show(label);

	dd.targetpath = patheditor_new(&core->targetpath);
	gtk_widget_show(dd.targetpath);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook),GTK_WIDGET(dd.targetpath),GTK_WIDGET(label));


	// Path to Targets

	label=gtk_label_new(_("Default options"));
	gtk_widget_show(label);

	// Table for aligned options.

	GtkWidget *table=gtk_table_new(2,5,FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table),8);
	gtk_table_set_col_spacing(GTK_TABLE(table),0,10);
	gtk_table_set_row_spacings(GTK_TABLE(table),8);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook),GTK_WIDGET(table),GTK_WIDGET(label));

	GtkAttachOptions gao = (GtkAttachOptions)(GTK_EXPAND|GTK_FILL);	

	// Argyll instrument selector

	label=gtk_label_new(_("Default instrument:"));
	gtk_table_attach(GTK_TABLE(table),label,0,1,0,1,GTK_SHRINK,GTK_SHRINK,0,0);

	// Build options for SimpleCombo
	SimpleComboOptions comboopts;
	ArgyllUISupport::AddInstrumentComboOpts(comboopts);

	// Build selector
	dd.instrument=simplecombo_new(comboopts);
	simplecombo_set(SIMPLECOMBO(dd.instrument),core->FindString("DefaultInstrument"));
	gtk_table_attach(GTK_TABLE(table),dd.instrument,1,2,0,1,gao,GTK_SHRINK,0,0);


	// Build resolution
	label=gtk_label_new(_("Default resolution:"));
	gtk_table_attach(GTK_TABLE(table),label,0,1,1,2,GTK_SHRINK,GTK_SHRINK,0,0);

	dd.resolution=gtk_spin_button_new_with_range(0.0,360.0,10.0);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(dd.resolution),core->FindInt("ChartResolution"));
	gtk_table_attach(GTK_TABLE(table),dd.resolution,1,2,1,2,gao,GTK_SHRINK,0,0);


	// Run the dialog

	gtk_widget_show_all(dialog);
	
	gint result=gtk_dialog_run(GTK_DIALOG(dialog));
	switch(result)
	{
		case GTK_RESPONSE_CANCEL:
			break;
		case GTK_RESPONSE_OK:
		case RESPONSE_SAVE:
			patheditor_get_paths(PATHEDITOR(dd.argyllpath),&argyll);
			char *agp=argyll.GetPaths();
			if(agp)
			{
				core->SetString("ArgyllPath",agp);
				free(agp);
			}
			patheditor_get_paths(PATHEDITOR(dd.targetpath),&core->targetpath);
			char *tgp=core->targetpath.GetPaths();
			if(tgp)
			{
				core->SetString("TargetPath",tgp);
				free(tgp);
			}
			core->SetString("DefaultInstrument",simplecombo_get(SIMPLECOMBO(dd.instrument)));
			core->SetInt("ChartResolution",gtk_spin_button_get_value(GTK_SPIN_BUTTON(dd.resolution)));
			if(result==RESPONSE_SAVE)
			{
//				gui.SaveGeometry();
				core->SaveConfigFile();
			}
			break;
	}
	gtk_widget_destroy(dialog);
}


//  Print Setup


struct printsetupdata
{
	LinProject *project;
	GtkWidget *optionbook;
	GtkWidget *printoutput;
	GtkWidget *dialog;
	stp_vars_t *backupvars;
	char *olddriver;
	stpui_optionbook_custompage custompage;
};


static void driver_changed(GtkWidget *wid,gpointer data)
{
	struct printsetupdata *cbd=(struct printsetupdata *)data;
	PrintOutput &po=cbd->project->printoutput;

	const char *driver=po.FindString("Driver");

	if(!(cbd->project->gprinter.SetDriver(po.FindString("Driver"))))
	{
		cerr << "Setting driver failed - reverting to default" << endl;
		po.SetString("Driver",DEFAULT_PRINTER_DRIVER);
		cbd->project->gprinter.SetDriver(po.FindString("Driver"));
		printoutputselector_refresh(PRINTOUTPUTSELECTOR(cbd->printoutput));
	}
	stpui_optionbook_rebuild(STPUI_OPTIONBOOK(cbd->optionbook));	
}


void PrintSetupDialog(ProjectGUI *project)
{
	printsetupdata dialogdata;

	dialogdata.project=project;
	dialogdata.backupvars=stp_vars_create_copy(project->gprinter.stpvars);
	dialogdata.olddriver=g_strdup(project->printoutput.FindString("Driver"));
	
	dialogdata.dialog=gtk_dialog_new_with_buttons(_("Printer Setup"),
		GTK_WINDOW(project->window),GtkDialogFlags(0),
		GTK_STOCK_CANCEL,GTK_RESPONSE_CANCEL,
		GTK_STOCK_OK,GTK_RESPONSE_OK,
		NULL);
	gtk_window_set_default_size(GTK_WINDOW(dialogdata.dialog),500,350);

	dialogdata.printoutput=printoutputselector_new(&project->printoutput);
	g_object_ref(G_OBJECT(dialogdata.printoutput));

	dialogdata.custompage.name=_("Output");
	dialogdata.custompage.widget=dialogdata.printoutput;
	
	dialogdata.optionbook=stpui_optionbook_new(project->gprinter.stpvars,&dialogdata.custompage,1);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialogdata.dialog)->vbox),dialogdata.optionbook,TRUE,TRUE,0);
	gtk_widget_show(dialogdata.optionbook);
	
	g_signal_connect(G_OBJECT(dialogdata.printoutput),"changed",G_CALLBACK(driver_changed),&dialogdata);

	gtk_widget_show(dialogdata.dialog);
	gint result=gtk_dialog_run(GTK_DIALOG(dialogdata.dialog));
	switch(result)
	{
		case GTK_RESPONSE_CANCEL:
			project->printoutput.DBToQueues();
			stp_vars_copy(project->gprinter.stpvars,dialogdata.backupvars);
			project->printoutput.SetString("Driver",dialogdata.olddriver);
			break;
		case GTK_RESPONSE_OK:
			project->printoutput.QueuesToDB();
//			const char *p1=stp_get_string_parameter(dialogdata.backupvars,"PageSize");
//			const char *p2=stp_get_string_parameter(state.printer.stpvars,"PageSize");
//			if(p1 && p2 && strcmp(p1,p2)!=0)
//				project->layout->UpdatePageSize();  // Needs to be run every time, but preserve margins.
			break;
	}
	g_object_unref(G_OBJECT(dialogdata.printoutput));
	stp_vars_destroy(dialogdata.backupvars);
	g_free(dialogdata.olddriver);
	gtk_widget_destroy(dialogdata.dialog);
}


