#ifndef DITHER_H
#define DITHER_H

#include "dropsize.h"

// FixMe - make this a template-based interface rather than a class heirarchy so we don't
// lose the ability to inline the decision function.
class DitherDropSizes
{
	public:
	DitherDropSizes(double smalldrop,double meddrop,double largedrop)
	{
		// We store the three dropsizes, and 0 to represented unprinted paper.
		drops[0]=0.0;
		drops[1]=smalldrop;
		drops[2]=meddrop;
		drops[3]=largedrop;
		// To simplify the PickDropSize function we compute a threshold for
		// each dropsize, half way between each size and the next.
		for(int i=0;i<3;++i)
			thresholds[i]=(drops[i]+drops[i+1])/2.0;
	}
	DitherDropSizes()
	{
		drops[0]=drops[1]=drops[2]=drops[3]=0.0;
		thresholds[0]=thresholds[1]=thresholds[2]=thresholds[3]=0.0;
	}
	virtual ~DitherDropSizes()
	{
	}
	virtual int PickDropSize(double input)
	{
		double r=rand()&31;
		r-=15.5;
//		input+=((1.0-input)*r)/220.0;
		input+=r/230.0;
		int i;
		for(i=0;i<3;++i)
		{
			if(input<thresholds[i])
				return(i);
		}
		return(i);
	}
	protected:
	double drops[4];
	double thresholds[3];
	friend class ImageSource_DropSizeDither_MultiDropSize;
};


// DitherDropSizes class for the simplest case -
// 1 dropsize, specified in constructor.
//  PickDropSize function returns 0 or 3

class DitherDropSizes_Single : public DitherDropSizes
{
	public:
	DitherDropSizes_Single(int dropsize=3) : DitherDropSizes(1.0,1.0,1.0), dropsize(dropsize)
	{
	}
	virtual ~DitherDropSizes_Single()
	{
	}
	virtual int PickDropSize(double input)
	{
		double r=rand()&63;
		r-=31.5;
//		input+=((1.0-input)*r)/70.0;
		input+=r/69.0;
		if(input<thresholds[0])
			return(0);
		return(dropsize);
	}
	int dropsize;
	friend class ImageSource_DropSizeDither_MultiDropSize;
};


// Adaptive DitherDropSizes class.
// Adjusts dropsize ratios from the Tuning target,
// based on input level.
// Given an input level for the largest dropsize, we
// find an L* value, and look up the input level at
// the other two dropsizes requred to match L*.

class DitherDropSizes_Adaptive : public DitherDropSizes
{
	public:
	DitherDropSizes_Adaptive(DropSizeResponse &d1,DropSizeResponse &d2,DropSizeResponse &d3)
		: DitherDropSizes(1.0,1.0,1.0), d1(d1), d2(d2), d3(d3)
	{
		lmin=0.0;
		lmax=100.0;
		double l;
		l=d1.InputToL(99.0);
		if(l>lmin)
			lmin=l;
		l=d1.InputToL(1.0);
		if(l<lmax)
			lmax=l;

		l=d2.InputToL(99.0);
		if(l>lmin)
			lmin=l;
		l=d2.InputToL(1.0);
		if(l<lmax)
			lmax=l;

		l=d3.InputToL(99.0);
		if(l>lmin)
			lmin=l;
		l=d3.InputToL(1.0);
		if(l<lmax)
			lmax=l;
		cerr << "Limiting L to " << lmin << " - " << lmax << endl;
	}
	virtual ~DitherDropSizes_Adaptive()
	{
	}
	virtual int PickDropSize(double input)
	{
		double r=rand()&31;
		r-=15.5;
		input+=r/230.0;

		if(input<0.0001)
			return(0);

		double gp3=input;

		// Find L* for large dropsize at gp_input
		// Then clamp to valid range

		double L=d3.GPToL(gp3);
		if(L<lmin)
		{
			L=lmin;
			gp3=d3.LToGP(L);
		}
		if(L>lmax)
		{
			L=lmax;
			gp3=d3.LToGP(L);
		}
		// Now find inputs for the other two dropsizes which yield the same L*
		double gp2=d2.LToGP(L);
		double gp1=d1.LToGP(L);

		// Set ratios...
		drops[1]=gp3/gp1;
		drops[2]=gp3/gp2;

		// ...and calculate thresholds on the fly.
		int i;
		for(i=0;i<3;++i)
		{
			if(input<((drops[i]+drops[i+1])/2.0))
				return(i);
		}
		return(i);
	}
	DropSizeResponse &d1;
	DropSizeResponse &d2;
	DropSizeResponse &d3;
	double lmin,lmax;
	friend class ImageSource_DropSizeDither_MultiDropSize;
};


// Use this class if you want to provide a separate dropsize selection strategy for each channel
// - i.e. small drops for one channel, large drops for another, etc.
// The dropsizes parameter must point to an array of pointers to DitherDropSizes with one entry for
// each channel.

class ImageSource_DropSizeDither_MultiDropSize : public ImageSource
{
	public:
	ImageSource_DropSizeDither_MultiDropSize(ImageSource *source,DitherDropSizes **dropsizes)
		: ImageSource(source), source(source), dropsizes(dropsizes), err1(NULL), err2(NULL)
	{
		err1=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		err2=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		for(int i=0;i<samplesperpixel*(width+2);++i)
			err1[i]=err2[i]=0; // (double(rand()%15)-7.5)/45.0;

		MakeRowBuffer();
		currentrow=-1;
	}
	virtual ~ImageSource_DropSizeDither_MultiDropSize()
	{
		if(source)
			delete source;
		if(err1)
			free(err1);
		if(err2)
			free(err2);
	}
	virtual ISDataType *GetRow(int row)
	{
		if(row==currentrow)
			return(rowbuffer);

		ISDataType *srcdata=source->GetRow(row);

		if(row & 1)
		{
			for(int s=0;s<samplesperpixel;++s)
				err2[s]=err2[samplesperpixel+s]=0;

			for(int x=0;x<width;++x)
			{
				for(int s=0;s<samplesperpixel;++s)
				{
					double t=srcdata[x*samplesperpixel+s];
					t/=IS_SAMPLEMAX;
					t+=err1[(x+1)*samplesperpixel+s];
					int d=dropsizes[s]->PickDropSize(t);
					double err=t-dropsizes[s]->drops[d];

					rowbuffer[x*samplesperpixel+s]=d;

					err1[(x+2)*samplesperpixel+s]+=(7.0*err)/16.0;
					err2[(x)*samplesperpixel+s]+=err/16.0;
					err2[(x+1)*samplesperpixel+s]+=(5.0*err)/16.0;
					err2[(x+2)*samplesperpixel+s]=(3.0*err)/16.0;
				}
			}
		}
		else
		{
			for(int s=0;s<samplesperpixel;++s)
				err1[width*samplesperpixel+s]=err1[(width+1)*samplesperpixel+s]=0;

			for(int x=width-1;x>=0;--x)
			{
				for(int s=0;s<samplesperpixel;++s)
				{
					double t=srcdata[x*samplesperpixel+s];
					t/=IS_SAMPLEMAX;
					t+=err2[(x+1)*samplesperpixel+s];
					int d=dropsizes[s]->PickDropSize(t);
					double err=t-dropsizes[s]->drops[d];

					rowbuffer[x*samplesperpixel+s]=d;

					err2[(x)*samplesperpixel+s]+=(7.0*err)/16.0;
					err1[(x)*samplesperpixel+s]=(3.0*err)/16.0;
					err1[(x+1)*samplesperpixel+s]+=(5.0*err)/16.0;
					err1[(x+2)*samplesperpixel+s]+=err/16.0;
				}
			}
		}
	
		currentrow=row;

		return(rowbuffer);
	}
	protected:
	ImageSource *source;
	DitherDropSizes **dropsizes;
	double *err1;
	double *err2;
	int mask;
};


// This is the most straightforward implementation - uses the same DitherDropSizes object
// for all channels.

class ImageSource_DropSizeDither : public ImageSource_DropSizeDither_MultiDropSize
{
	public:
	ImageSource_DropSizeDither(ImageSource *source,DitherDropSizes &dropsize)
		: ImageSource_DropSizeDither_MultiDropSize(source,NULL)
	{
		dropsizes=new DitherDropSizes *[samplesperpixel];
		for(int i=0;i<samplesperpixel;++i)
			dropsizes[i]=&dropsize;
		Init();
	}
	ImageSource_DropSizeDither(ImageSource *source,double ds1,double ds2,double ds3)
		: ImageSource_DropSizeDither_MultiDropSize(source,NULL)
	{
		DitherDropSizes sizes(ds1,ds2,ds3);
		ditherdropsizes=sizes;

		cerr << "Ds1: " << ds1 << ", ds2: " << ds2 << ", ds3: " << ds3 << endl;

		dropsizes=new DitherDropSizes *[samplesperpixel];
		for(int i=0;i<samplesperpixel;++i)
			dropsizes[i]=&ditherdropsizes;
		Init();
	}
	void Init()
	{
		// This stuff is taken care of by the superclass
#if 0
		err1=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		err2=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		for(int i=0;i<samplesperpixel*(width+2);++i)
			err1[i]=err2[i]=0; // (double(rand()%15)-7.5)/45.0;

		MakeRowBuffer();
		currentrow=-1;
#endif
	}
	virtual ~ImageSource_DropSizeDither()
	{
		if(dropsizes)
			delete[] dropsizes;
	}
	protected:
	// Used if caller provides actual ratios rather than a preformed DitherDropSizes
	DitherDropSizes ditherdropsizes;
};


// A special-purpose dither class used as an equivalent to Gutenprint's "drop-size segmented" dither.
// The channels will be processed in threes - small, medium, large, small, medium... etc.

class ImageSource_DropSizeSeparatedDither : public ImageSource_DropSizeDither_MultiDropSize
{
	public:
	ImageSource_DropSizeSeparatedDither(ImageSource *source)
		: ImageSource_DropSizeDither_MultiDropSize(source,NULL)
	{
		dss[0]=DitherDropSizes_Single(1);
		dss[1]=DitherDropSizes_Single(2);
		dss[2]=DitherDropSizes_Single(3);

		dropsizes=new DitherDropSizes *[samplesperpixel];
		for(int i=0;i<samplesperpixel;++i)
			dropsizes[i]=&dss[i%3];
		Init();
	}
	void Init()
	{
		err1=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		err2=(double *)malloc(sizeof(double)*samplesperpixel*(width+2));
		for(int i=0;i<samplesperpixel*(width+2);++i)
			err1[i]=err2[i]=0; // (double(rand()%15)-7.5)/45.0;

		MakeRowBuffer();
		currentrow=-1;
	}
	virtual ~ImageSource_DropSizeSeparatedDither()
	{
		if(dropsizes)
			delete[] dropsizes;
	}
	protected:
	// Used if caller provides actual ratios rather than a preformed DitherDropSizes
	DitherDropSizes_Single dss[3];
};

#endif

