#include <iostream>
#include <fstream>
#include <cstring>

#include <getopt.h>

#include <gtk/gtk.h>

using namespace std;

#include "support/util.h"
#include "channelresponse.h"
#include "linearization.h"
#include "linproject.h"
#include "gpdevicensupport.h"
#include "ticgats.h"

#include "imagesource/imagesource_modifiedgamma.h"


class ExportOpts
{
	public:
	ExportOpts(int argc,char *argv[]) : applydensity(true), applygamma(true), applyinklimit(false), steps(101), outputfilename(NULL)
	{
		static struct option long_options[] =
		{
			{"help",no_argument,NULL,'h'},
			{"version",no_argument,NULL,'v'},
			{"ignoredensity",no_argument,NULL,'d'},
			{"ignoregamma",no_argument,NULL,'g'},
			{"applyinklimit",no_argument,NULL,'i'},
			{0, 0, 0, 0}
		};

		while(1)
		{
			int c;
			c = getopt_long(argc,argv,"hvdgis:o:",long_options,NULL);
			if(c==-1)
				break;
			switch (c)
			{
				case 'h':
					Usage(argc,argv);
					throw 0;
					break;
				case 'v':
					cerr << "ExportLin - part of " << PACKAGE_STRING << endl;
					throw 0;
					break;
				case 'd':
					applydensity=false;
					break;
				case 'g':
					applygamma=false;
					break;
				case 'i':
					applyinklimit=true;
					break;
				case 's':
					steps=atoi(optarg);
					break;
				case 'o':
					outputfilename=optarg;
					break;
			}
		}
		if(optind>=argc)
		{
			Usage(argc,argv);
			throw 5;
		}
	}
	~ExportOpts()
	{
	}
	void Usage(int argc,char *argv[])
	{
		cerr << "Usage: " << argv[0] << " [options] ProjectName.gplin" << endl;
		cerr << "\t -h --help\t\tdisplay this message" << endl;
		cerr << "\t -v --version\t\tdisplay version" << endl;
		cerr << "\t -s --steps\t\tThe number of steps in the output curves - default 101" << endl;
		cerr << "\t -d --ignoredensity\tDon't apply density correction to the linearization" << endl;
		cerr << "\t -g --ignoregamma\tDon't apply gamma correction to the linearization" << endl;
		cerr << "\t -i --inklimit\tDon't scale each channel by its inklimit" << endl;
		cerr << "\t -o --outputfile\tOutput filename for CSV file containing the linearization curves" << endl;
	}
	bool applydensity;
	bool applygamma;
	bool applyinklimit;
	int steps;
	const char *outputfilename;
};


int main(int argc,char **argv)
{
	gtk_init_check(&argc,&argv);
	try
	{
		ExportOpts opts(argc,argv);

		LinearizationCore core;
		LinProject proj(core);
		proj.SetFilename(argv[optind]);
		proj.Load();

		char *ti3name=proj.GetTempRoot("RawLin",".ti3");
		proj.gprinter.SetCurvesFromTI3(ti3name);
		Linearization &lin=proj.gprinter.linearization;

		RawChannelResponses rcr(ti3name);
		free(ti3name);

		ChannelResponse *c=rcr.FirstChannel();
		int chan=0;
		while(c)
		{
			cerr << "Setting inklimit of channel " << chan << " to " << lin.InkLimitFromDB(chan) << endl;
			c->SetInkLimit(lin.InkLimitFromDB(chan));
			c=c->NextChannel();
			++chan;
		}

		ostream *outstream=&cout;
		fstream outfile;
		if(opts.outputfilename)
		{
			cerr << "Opening file: " << opts.outputfilename << endl;
			outfile.open(opts.outputfilename,ios_base::out);
			outstream=&outfile;
		}

		*outstream << "Input";
		ChannelResponse *c1=rcr.FirstChannel();
		while(c1)
		{
			*outstream << ", " << c1->GetShortName();
			c1=c1->NextChannel();
		}
		*outstream << endl;

		double bdens=lin.FindFloat("BaseDensity");

		for(double i=0;i<opts.steps;++i)
		{
			int chan=0;
			double di=i;
			di*=100.0/(opts.steps-1);
			*outstream << di/100.0;
			ChannelResponse *c1=rcr.FirstChannel();
			while(c1)
			{
				double inklimit=c1->GetInkLimit()/100.0;
				double out=c1->ReverseInterpolate(di)/100.0;
				if(out<0.0) out=0.0;
				if(out>1.0) out=1.0;

				if(opts.applygamma)
					out=lin.ForwardGamma(out);

				if(opts.applydensity)
					out*=bdens;

				if(opts.applyinklimit)
					out/=inklimit;

				*outstream << ", " << out;

				c1=c1->NextChannel();
				++chan;
			}
			*outstream << endl;
		}
	}
	catch(const char *err)
	{
		cerr << "Error: " << err << endl;
	}
	catch(int rc)
	{
		return(rc);
	}
	return(0);
}

