#ifndef IMAGESOURCE_DOTSCREEN_H
#define IMAGESOURCE_DOTSCREEN_H

#include <iostream>
#include <cmath>

#include "imagesource.h"

class SpotMatrix_Old
{
	public:
	SpotMatrix_Old(int size) : size(size)
	{
		values=new double[size*size];

		double centre=double(size-1)/2;
		double shellwidth=1.0/size;
		double diag=(sqrt(2)*centre) / sqrt((1-shellwidth)*(1-shellwidth));

		std::cout << "Shellwidth " << shellwidth << ", diag " << diag << std::endl;

		for(int y=0;y<size;++y)
		{
			double cy=y-centre;
			for(int x=0;x<size;++x)
			{
				double cx=x-centre;
				double r=sqrt(cx*cx+cy*cy);
//				std::cout << cx << ", " << cy << " -> " << r << ", ";
				double spiral=atan2(cy,cx);
//				std::cout << spiral << ", ";
				if(spiral<0) spiral+=2*M_PI;
				spiral=spiral*shellwidth/(2*M_PI);
//				std::cout << spiral << ", ";
				r=r/diag+spiral;
				values[y*size+x]=r;
//				std::cout << r << std::endl;
			}
		}

		double min=2, max=0;
		for(int y=0;y<size;++y)
		{
			for(int x=0;x<size;++x)
			{
				if(Value(x,y)<min)
					min=Value(x,y);
				if(Value(x,y)>max)
					max=Value(x,y);
			}
		}
//		std::cout << "Min " << min << ", " << max << std::endl;

		double tmin=1.0/(size*size);
		double tmax=1.0-tmin;
		double trange=tmax-tmin;
		double range=max-min;
		for(int y=0;y<size;++y)
		{
			for(int x=0;x<size;++x)
			{
				double v=(Value(x,y)-min)*trange/range+tmin;
				Value(x,y)=v;
			}
		}
	}
	~SpotMatrix_Old()
	{
		if(values)
			delete[] values;
	}
	inline double &Value(int x,int y)
	{
		return(values[y*size+x]);
	}	
	protected:
	int size;
	double *values;
};



class SpotMatrix
{
	public:
	SpotMatrix(int size) : size(size)
	{
		values=new ISDataType[size*size];
		double step=IS_SAMPLEMAX;
		step/=(size*size)+0.5;
		double current=step;

		for(int y=0;y<size;++y)
		{
			for(int x=0;x<size;++x)
			{
				Value(x,y)=0;
			}
		}

		double maxr=sqrt(size*size+size*size);
		for(double r=0.5;r<maxr;r+=0.5)
		{
			double astep=atan2(1.0,r);
//			cerr << "astep: " << astep << " from " << r << endl;
			for(double a=0;a<2*M_PI;a+=astep)
			{
//				cerr << "Angle " << a << ", astep: " << astep << " from " << r << endl;
				double x=size/2.0+r*cos(a);
				double y=size/2.0+r*sin(a);
				if(x>=0 && x<size && y>=0 && y<size)
				{
//					std::cout << a << " - > " << x << ", " << y << endl;
					if(Value(int(x),int(y))==0)
					{
						Value(int(x),int(y))=current;
						current+=step;
					}
				}
			}
		}

#if 0
		for(int y=0;y<size;++y)
		{
			for(int x=0;x<size;++x)
			{
				std::cout << Value(x,y) << "\t";
			}
			std::cout << std::endl;
		}
#endif
	}
	~SpotMatrix()
	{
		if(values)
			delete[] values;
	}
	inline ISDataType &Value(int x,int y)
	{
		return(values[y*size+x]);
	}	
	protected:
	int size;
	ISDataType *values;
};



class ScreenParam
{
	public:
	ScreenParam(double angle,double freq) : angle(angle), freq(freq)
	{
	}
	ScreenParam() : angle(0.0), freq(0.0)
	{
	}
	~ScreenParam()
	{
	}
	double angle;
	double freq;
};


#define GLOBALSPOTMATRIXSIZE 512
extern SpotMatrix GlobalSpotMatrix;

class ImageSource_DotScreen : public ImageSource
{
	public:
	ImageSource_DotScreen(ImageSource *is,ScreenParam *screenparams) : ImageSource(is), source(is)
	{
		randomaccess=false;
		counters=new DotScreen_Counter[samplesperpixel];
		for(int i=0;i<samplesperpixel;++i)
		{
			counters[i]=DotScreen_Counter(screenparams[i],this);
		}
		err=(int *)malloc(sizeof(int)*samplesperpixel*width);
		for(int i=0;i<width*samplesperpixel;++i)
			err[i]=0;
		MakeRowBuffer();
	}
	~ImageSource_DotScreen()
	{
		if(counters)
			delete[] counters;
	}
	ISDataType *GetRow(int row)
	{
		int rerr[samplesperpixel];
		for(int i=0;i<samplesperpixel;++i)
			rerr[i]=0;

		if(currentrow==row)
			return(rowbuffer);
		for(int i=0;i<samplesperpixel;++i)
			counters[i].NextRow();

//		std::cout << "Dotscreen getting row " << row << std::endl;
		ISDataType *src=source->GetRow(row);

		for(int x=0;x<width;++x)
		{
			for(int i=0;i<samplesperpixel;++i)
			{
				ISDataType t=counters[i].NextPixel();
				int p=src[x*samplesperpixel+i];
				p+=(err[x*samplesperpixel+i]+rerr[i])/16;
				if(p>t)
					t=IS_SAMPLEMAX;
				else
					t=0;
				rowbuffer[x*samplesperpixel+i]=t;
				err[x*samplesperpixel+i]=p-t;
			}
		}

		currentrow=row;
		return(rowbuffer);
	}
	protected:
	class DotScreen_Counter
	{
		public:
		DotScreen_Counter(ScreenParam &screenparam,const ImageSource *source) : param(screenparam)
		{
			rowx=rowy=0;
			curx=cury=0;
			double step=param.freq/source->xres;
			incx=step*cos((M_PI*param.angle)/180);
			incy=step*sin((M_PI*param.angle)/180);
			step=param.freq/source->yres;
			vincx=-step*sin((M_PI*param.angle)/180);
			vincy=step*cos((M_PI*param.angle)/180);
		}
		DotScreen_Counter()
		{
		}
		~DotScreen_Counter()
		{
		}
		void NextRow()
		{
			curx=rowx+=vincx;
			cury=rowy+=vincy;
		}
		inline double NextPixel()
		{
			curx+=incx;
			cury+=incy;
			double tx=curx-int(curx);
			double ty=cury-int(cury);
			if(tx<0) tx+=1.0;
			if(ty<0) ty+=1.0;
			return(GlobalSpotMatrix.Value(int(tx*GLOBALSPOTMATRIXSIZE), int(ty*GLOBALSPOTMATRIXSIZE)));
		}
		protected:
		ScreenParam param;
		double rowx,rowy;
		double curx,cury;
		double incx,incy;
		double vincx,vincy;
	};
	ImageSource *source;	
	DotScreen_Counter *counters;
	int *err;
};

#endif

