#include <iostream>

#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_modifiedgamma.h"
#include "imagesource/imagesource_gamma.h"

#include "linearization.h"

using namespace std;


// Experimental function - will not be used.
//
// ImageSource for normalizing the total area coverage to 1.0
// This is just a prototype for evaluating the improvement in gamut
// yeilded by this technique - a "proper" implementation would have
// to determine the maximum output level that would correspond to
// a given fraction of the current input level's ink coverage, and
// be performed *before* the linearization.  This prototype will be
// wedged in *after* the linearization, which will compromise the
// linearization for colours with more than one colorant in use.
// The algorithm is as follows:
// out[i]=in[i]/(sum(in[])*1.0/max(in[]);

class ImageSource_Normalize : public ImageSource
{
	public:
	ImageSource_Normalize(ImageSource *source)
		: ImageSource(source), source(source)
	{
		MakeRowBuffer();
	}

	~ImageSource_Normalize()
	{
		if(source)
			delete source;
	}

	ISDataType *GetRow(int row)
	{
		ISDataType *src;

		if(row==currentrow)
			return(rowbuffer);

		src=source->GetRow(row);

		if(HAS_ALPHA(type))
		{
			for(int i=0;i<width;++i)
			{
				float a=0.0;
				int max=0;
				int j;
				for(j=0;j<samplesperpixel-1;++j)
				{
					int t=src[i*samplesperpixel+j];
					a+=t;
					if(max<t)
						max=t;
				}
				if(max)
					a=max/a;
				else
					a=1.0;

				for(j=0;j<samplesperpixel-1;++j)
					rowbuffer[i*samplesperpixel+j]=ISDataType(src[i*samplesperpixel+j]*a);
				rowbuffer[i*samplesperpixel+j]=src[i*samplesperpixel+j];	// Copy alpha unchanged
			}	
		}
		else
		{
			for(int i=0;i<width;++i)
			{
				float a=0.0;
				int max=0;
				int j;
				for(j=0;j<samplesperpixel;++j)
				{
					int t=src[i*samplesperpixel+j];
					a+=t;
					if(max<t)
						max=t;
				}
				if(max)
					a=max/a;
				else
					a=1.0;

				for(j=0;j<samplesperpixel;++j)
					rowbuffer[i*samplesperpixel+j]=ISDataType(src[i*samplesperpixel+j]*a);
			}	
		}
		currentrow=row;
		return(rowbuffer);
	}
	protected:
	ImageSource *source;
};


// ImageSource for applying curves

class ImageSource_CorrectionCurves : public ImageSource
{
	public:
	ImageSource_CorrectionCurves(ImageSource *source,CorrectionCurve *curves)
		: ImageSource(source), source(source), curves(curves)
	{
		MakeRowBuffer();
	}

	~ImageSource_CorrectionCurves()
	{
		if(source)
			delete source;
	}

	ISDataType *GetRow(int row)
	{
		ISDataType *src;

		if(row==currentrow)
			return(rowbuffer);

		src=source->GetRow(row);

		if(HAS_ALPHA(type))
		{
			for(int i=0;i<width;++i)
			{
				int j;
				for(j=0;j<samplesperpixel-1;++j)
					rowbuffer[i*samplesperpixel+j]=curves[j][src[i*samplesperpixel+j]];
				rowbuffer[i*samplesperpixel+j]=src[i*samplesperpixel+j];	// Copy alpha unchanged
			}	
		}
		else
		{
			for(int i=0;i<width;++i)
			{
				int j;
				for(j=0;j<samplesperpixel;++j)
					rowbuffer[i*samplesperpixel+j]=curves[j][src[i*samplesperpixel+j]];
			}	
		}
		currentrow=row;
		return(rowbuffer);
	}
	protected:
	ImageSource *source;
	CorrectionCurve *curves;
};


// Linearization class

ConfigTemplate Linearization::Template[]=
{
	ConfigTemplate("BaseDensity",float(1.0)),
	ConfigTemplate("BaseGamma",float(1.0)),
	ConfigTemplate("TotalInkLimit",float(270)),
	ConfigTemplate("ChannelInkLimit0",float(100.0)),
	ConfigTemplate("ChannelInkLimit1",float(100.0)),
	ConfigTemplate("ChannelInkLimit2",float(100.0)),
	ConfigTemplate("ChannelInkLimit3",float(100.0)),
	ConfigTemplate("ChannelInkLimit4",float(100.0)),
	ConfigTemplate("ChannelInkLimit5",float(100.0)),
	ConfigTemplate("ChannelInkLimit6",float(100.0)),
	ConfigTemplate("ChannelInkLimit7",float(100.0)),
	ConfigTemplate()
};


Linearization::Linearization(ConfigFile *file) : ConfigDB(Template)
{
	new ConfigDBHandler(file,"[Linearization]",this);
}


Linearization::~Linearization()
{

}


ImageSource *Linearization::Apply(ImageSource *source)
{

	// Apply curves here...
	source=new ImageSource_CorrectionCurves(source,curves);

	source=new ImageSource_ModifiedGamma(source,FindFloat("BaseGamma"));
//	source=new ImageSource_Gamma(source,FindFloat("BaseGamma"));

	ISDeviceNValue density(source->samplesperpixel,ISDataType(FindFloat("BaseDensity")*IS_SAMPLEMAX));
	source=new ImageSource_ScaleDensity(source,density);


	// Apply temporary normalization...

//	source=new ImageSource_Normalize(source);


	return(source);
}


double Linearization::ForwardGamma(double in)
{
	double gam=FindFloat("BaseGamma");
	return(1.0-ImageSource_ModifiedGamma::ModifiedGamma(1.0-in,gam));
}


double Linearization::ReverseGamma(double in)
{
	double gam=FindFloat("BaseGamma");
	cerr << "Reverse correcting " << 1.0-in << " by gamma " << gam << endl;
	cerr << "Inverse modified gamma:" << ImageSource_ModifiedGamma::InverseModifiedGamma(1.0-in,gam) << endl;
	return(1.0-ImageSource_ModifiedGamma::InverseModifiedGamma(1.0-in,gam));
}


CorrectionCurve &Linearization::operator[](int i)
{
	if(i<0 || i>=LINEARIZATION_MAX_CHANNELS)
		throw "Linearization - Curve index out of range!";
	return(curves[i]);
}


// The ConfigDB doesn't support arrays yet - so we just use an array of key names.
static const char *channellimits[]=
{
	"ChannelInkLimit0",
	"ChannelInkLimit1",
	"ChannelInkLimit2",
	"ChannelInkLimit3",
	"ChannelInkLimit4",
	"ChannelInkLimit5",
	"ChannelInkLimit6",
	"ChannelInkLimit7",
};


void Linearization::InkLimitToDB(int chan,float limit)
{
	if(chan>7)
		throw "Linearization currently supports maximum 8 channels.";
	SetFloat(channellimits[chan],limit);
}


float Linearization::InkLimitFromDB(int chan)
{
	if(chan>7)
		throw "Linearization currently supports maximum 8 channels.";
	return(FindFloat(channellimits[chan]));
}

