#include <iostream>

#include "imagesource/imagesource_modifiedgamma.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_devicen_remap.h"
#include "support/util.h"

#include "imagesource_dropsizemerge.h"
#include "imagesource_dotscreen.h"

#include "linearizedprinter.h"
#include "correctioncurve.h"
#include "channelresponse.h"
#include "gpdevicensupport.h"
#include "dropsize.h"
#include "dither.h"

using namespace std;

// Subclass of CorrectionCurve - used within SetCurvesFromTI3()

class CorrectionCurve_ChannelResponse : public CorrectionCurve
{
	public:
	CorrectionCurve_ChannelResponse(ChannelResponse *c) : CorrectionCurve()
	{
		for(int i=0;i<=IS_SAMPLEMAX;++i)
		{
			double in=(i*100.0)/IS_SAMPLEMAX;
			double out=c->ReverseInterpolate(in);
			if(out<0.0) out=0.0;
			if(out>100.0) out=100.0;
			ISDataType o=ISDataType((IS_SAMPLEMAX*out)/100.0);
			correction[i]=o;
//			cerr << "Mapping " << i << " to " << o << endl;
		}
	}
	protected:
};



LinearizedPrinter::LinearizedPrinter(PrintOutput &output,ConfigFile *inf,const char *section)
	: GPrinter(output,inf,section), linearization(inf), tuning(inf)
{
}


LinearizedPrinter::~LinearizedPrinter()
{
}


void LinearizedPrinter::Print(ImageSource *source,int xpos,int ypos)
{
	cerr << "*** LinearizedPrinter: Printing at position: " << xpos << ", " << ypos << endl;
	source=linearization.Apply(source);
	GPrinter::Print(source,xpos,ypos);
}


void LinearizedPrinter::PrintPreDithered(ImageSource *source,int xpos,int ypos)
{
	cerr << "*** LinearizedPrinter: Printing pre-dithered: " << xpos << ", " << ypos << endl;
	char *olddither=SafeStrdup(stp_get_string_parameter(stpvars,"DitherAlgorithm"));
	stp_set_string_parameter(stpvars,"DitherAlgorithm","Predithered");
	GPrinter::Print(source,xpos,ypos);
	stp_set_string_parameter(stpvars,"DitherAlgorithm",olddither);
	free(olddither);
}


void LinearizedPrinter::SetCurvesFromTI3(const char *fn)
{
	GPColorantList colorants(this);

	RawChannelResponses resp(fn);

	int *channelmap=resp.CreateRawChannelMap(colorants);

	int chan=0;
	ChannelResponse *c=resp.FirstChannel();
	while(c)
	{
		int channelidx=channelmap[resp.GetColorantIndex(c->GetLongName())];
		cerr << "Assigning channel " << c->GetName() << " (" << c->GetLongName() << ")" << endl;
		cerr << "Channel index is: " << colorants.GetColorantIndex(c->GetLongName()) << endl;
		cerr << "Channel index is: " << channelidx << endl;

		float inklimit=100.0;
		try
		{
			inklimit=linearization.InkLimitFromDB(chan);
		}
		catch (const char *err)
		{
			cerr << err << endl;
		}

		cerr << "Setting ink limit to:" << inklimit << endl;
		c->SetInkLimit(inklimit);

		CorrectionCurve_ChannelResponse temp(c);
		linearization[channelidx]=temp;
		c=c->NextChannel();
		++chan;
	}
	delete[] channelmap;
}


// FIXME - maybe a better fit in the Tuning object?

void LinearizedPrinter::SetDropSizesFromTI3(const char *fn)
{
	RawChannelResponses ti3(fn);

	ChannelResponse *c=NULL;

	float gamma=tuning.FindFloat("TuningGamma");
	cerr << "Using gamma: " << gamma << endl;
	c=ti3.GetChannel("Light Light Black");
	DropSizeResponse d1(c,tuning.FindFloat("DropSizeLimit0"),gamma);

	c=ti3.GetChannel("Light Black");
	DropSizeResponse d2(c,tuning.FindFloat("DropSizeLimit1"),gamma);

	c=ti3.GetChannel("Black");
	DropSizeResponse d3(c,tuning.FindFloat("DropSizeLimit2"),gamma);

	double input=90;
	double L=d1.InputToL(input);
	double gp1=d1.InputToGP(input);
	cerr << "Input: " << input << " (" << gp1 << ") -> " << L << endl;
	double gp2=d2.LToGP(L);
	double gp3=d3.LToGP(L);
	cerr << " -> " << gp2 << "  -  Ratio 1: " << gp3/gp2 << endl;
	cerr << " -> " << gp3 << "  -  Ratio 2: " << gp3/gp1 << endl;

	tuning.DropSizeRatioToDB(0,gp3/gp1);
	tuning.DropSizeRatioToDB(1,gp3/gp2);
	tuning.DropSizeRatioToDB(2,1.0);
}


ImageSource *LinearizedPrinter::ScaleToPrinterResolution(ImageSource *source,IS_ScalingQuality quality)
{
	int xres,yres;
	stp_describe_resolution(stpvars,&xres,&yres);
	return(ISScaleImageByResolution(source,xres,yres,quality));
}


ImageSource *LinearizedPrinter::Dither(ImageSource *is)
{
	is=new ImageSource_ModifiedGamma(is,tuning.FindFloat("TuningGamma"));
	ISDeviceNValue densities(is->samplesperpixel,0);
	for(int s=0;s<is->samplesperpixel;++s)
		densities[s]=IS_SAMPLEMAX*tuning.FindFloat("DropSizeLimit2");
	is=new ImageSource_ScaleDensity(is,densities);

	is=ScaleToPrinterResolution(is);
	double ds1=tuning.DropSizeRatioFromDB(0);
	double ds2=tuning.DropSizeRatioFromDB(1);
	double ds3=tuning.DropSizeRatioFromDB(2);
	is=new ImageSource_DropSizeDither(is,ds1,ds2,ds3);
	return(is);
}


class ImageSource_ChannelToDropSize : public ImageSource
{
	public:
	ImageSource_ChannelToDropSize(ImageSource *src) : ImageSource(src), source(src)
	{
		samplesperpixel=(source->samplesperpixel+2)/3;
		MakeRowBuffer();
	}
	~ImageSource_ChannelToDropSize()
	{
		if(source)
			delete source;
	}
	ISDataType *GetRow(int row)
	{
		if(row==currentrow)
			return(rowbuffer);
		ISDataType *src=source->GetRow(row);
		for(int x=0;x<width;++x)
		{
			for(int s=0;s<samplesperpixel;++s)
				rowbuffer[x*samplesperpixel+s]=0;

			for(int s=0;s<source->samplesperpixel;++s)
			{
				int s2=s/3;
				if(src[x*source->samplesperpixel+s] && !rowbuffer[x*samplesperpixel+s2])
					rowbuffer[x*samplesperpixel+s2]=(s%3)+1;
			}
		}
		return(rowbuffer);
	}
	protected:
	ImageSource *source;
};


ImageSource *LinearizedPrinter::DropSizeSeparatedDither(ImageSource *is)
{
	is=new ImageSource_ModifiedGamma(is,tuning.FindFloat("TuningGamma"));
	double dsl[3];
	dsl[0]=IS_SAMPLEMAX*tuning.FindFloat("DropSizeLimit0");
	dsl[1]=IS_SAMPLEMAX*tuning.FindFloat("DropSizeLimit1");
	dsl[2]=IS_SAMPLEMAX*tuning.FindFloat("DropSizeLimit2");

	ISDeviceNValue densities(is->samplesperpixel,0);
	for(int s=0;s<is->samplesperpixel;++s)
		densities[s]=dsl[s%3];
	is=new ImageSource_ScaleDensity(is,densities);

	std::cerr << "Scaling to printer resolution..." << endl;

	is=ScaleToPrinterResolution(is);

#if 1
//  Try using a dot screen here.
	ScreenParam params[]=
	{
		ScreenParam(15,100),
		ScreenParam(15,100),
		ScreenParam(15,100),
		ScreenParam(15,100),
	};
	std::cerr << "Defined dotscreens - applying..." << endl;
	is=new ImageSource_DotScreen(is,params);
	std::cerr << "Merging channels to dropsizes..." << endl;

	is=new ImageSource_ChannelToDropSize(is);
#endif

#if 0
//  An over-complicated Floyd-Steinberg implementation!
	is=new ImageSource_DropSizeSeparatedDither(is);
	is=new ImageSource_DropSizeMerge(is);
#endif

	// Remap to printer DeviceN
	GPColorantList gpcolorants(this);
	int colcount=gpcolorants.GetColorantCount();
	int blackidx=gpcolorants.GetColorantIndex("Black");
	if(blackidx<0)
		blackidx=gpcolorants.GetColorantIndex("Photo Black");
	cerr << "Got black index:" << blackidx << endl;
	int maptable[2]={0,0};
	maptable[0]=blackidx;
	int magidx=gpcolorants.GetColorantIndex("Magenta");
	if(magidx<0)
		magidx=gpcolorants.GetColorantIndex("Vivid Magenta");
	cerr << "Got magenta index:" << magidx << endl;
	maptable[1]=magidx;

	is=new ImageSource_DeviceN_Remap(is,maptable,colcount);

	return(is);
}

