#include <iostream>
#include <fstream>

#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <libgen.h>

#include "support/searchpath.h"
#include "support/util.h"

#include "imagesource/imagesource_devicen_remap.h"

#include "gpdevicensupport.h"
#include "linproject.h"


#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)

using namespace std;


ConfigTemplate LinProject::Template[]=
{
	ConfigTemplate("ChartSeed",int(0)),	// Random number used when generating charts.
										// Storing this number makes chart generation repeatable.
	ConfigTemplate("Instrument","DTP41"),
	ConfigTemplate("StepsPerChannel",int(64)),
	ConfigTemplate("Colorants",""),
	ConfigTemplate("DropsizeStepsPerChannel",int(64)),
	ConfigTemplate("DropsizeColorants",""),
	ConfigTemplate()
};


LinProject::LinProject(LinearizationCore &core,const char *fn)
	: ConfigFile(), ConfigDB(Template), printoutput(this,"[Output]"), gprinter(printoutput,this,"[Print]"),
	core(core), filename(NULL), tempdirname(NULL), temproot(NULL), projecttype(LINPROJECTTYPE_LINEARIZATION)
{
	new ConfigDBHandler(this,"[GPLin]",this);
	SetInt("ChartSeed",RandomSeeded(100));
	cerr << "Set chart seed to " << FindInt("ChartSeed") << endl;
	SetString("Instrument",core.FindString("DefaultInstrument"));
	if(fn)
	{
		SetFilename(fn);
		Load();
	}
}


void LinProject::SetFilename(const char *fn)
{
	if(filename)
		free(filename);
	if(tempdirname)
		free(tempdirname);
	if(temproot)
		free(temproot);
	filename=strdup(fn);
	tempdirname=BuildFilename(filename,"_temp",NULL);

	char *tmp=BuildFilename(filename,"",NULL);
	char *base=basename(tmp);
	temproot=(char *)malloc(strlen(base)+strlen(tempdirname)+4);
	sprintf(temproot,"%s%c%s",tempdirname,SEARCHPATH_SEPARATOR,base);
	free(tmp);
}


const char *LinProject::GetFilename()
{
	return(filename);
}


void LinProject::MakeTempDir()
{
	if(!CreateDirIfNeeded(tempdirname))
		throw _("Can't create directory for temporary files!");
}


const char *LinProject::GetTempDir()
{
	return(tempdirname);
}


char *LinProject::GetTempRoot(const char *id,const char *ext)
{
	if(!ext)
		ext="";
	if(!temproot)
		return(NULL);
	int n=strlen(temproot)+strlen(id)+strlen(ext)+2;
	char *result=(char *)malloc(n);
	sprintf(result,"%s_%s%s",temproot,id,ext);
	return(result);
}


LinProject::~LinProject()
{
	if(filename)
		free(filename);
	if(tempdirname)
		free(tempdirname);
	if(temproot)
		free(temproot);
}


bool LinProject::Load()
{
	bool result=false;
	if(filename)
	{
		result=ParseConfigFile(filename);
		printoutput.DBToQueues();
	}
	return(result);
}


bool LinProject::Save()
{
	printoutput.QueuesToDB();
	if(filename)
		return(SaveConfigFile(filename));
	return(false);
}


bool LinProject::Import(const char *file)
{
	bool result=false;
	if(file)
	{
		result=ParseConfigFile(file);
		printoutput.DBToQueues();
	}
	return(result);
}


void LinProject::CreateRawChannelLinChart(const char *id,Progress *p)
{
	CreateDirIfNeeded(tempdirname);
	cerr << "Generating test target..." << endl;
	GPColorantList colorants(&gprinter);
	colorants.SetEnabledColorants(FindString("Colorants"));

	char *fnroot=GetTempRoot(id);
	TIXFile *ti1=core.CreateRawChannelLinChart(fnroot,&colorants,FindInt("StepsPerChannel"),p);
	ArgyllInstrument inst(FindString("Instrument"));

	gprinter.GetImageableArea();
	float pw=gprinter.pagewidth;
	float ph=gprinter.pageheight;
	cerr << "Using pagesize: " << pw << ", " << ph << endl;
	pw=(pw*25.4)/72;
	ph=(ph*25.4)/72;
	ArgyllPaperSize ps(int(pw+0.5),int(ph+0.5));
	int res=core.FindInt("ChartResolution");
	Argyll_TestChart_TIFFs *tiffs=core.RenderTIFFs(fnroot,ps,inst,res,FindInt("ChartSeed"),p);
	delete tiffs;
	delete ti1;
	free(fnroot);
}


void LinProject::CreateDropSizeLinChart(const char *id,Progress *p)
{
	CreateDirIfNeeded(tempdirname);
	cerr << "Generating dropsize target..." << endl;
//	DropSize_Colorants colorants(&gprinter);

	// Copy existing target to the project directory.
	char *targetsrc=core.targetpath.SearchPaths("DropSize.ti1");
	if(!targetsrc)
		throw "DropSize Target not found!\nPlease either install the targets from the GPLin distribution\nor use the preferences dialog to locate them.";

	cerr << "Found: " << targetsrc << endl;
	char *outfn=GetTempRoot("DropSize",".ti1");
	cerr << "Output filename: " << outfn << endl;

	ifstream in(targetsrc);
	ofstream out(outfn);
	out << in.rdbuf();
	in.close();
	out.close();
	free(outfn);
	free(targetsrc);

	char *fnroot=GetTempRoot(id);

	ArgyllInstrument inst(FindString("Instrument"));

	gprinter.GetImageableArea();
	float pw=gprinter.pagewidth;
	float ph=gprinter.pageheight;
	cerr << "Using pagesize: " << pw << ", " << ph << endl;
	pw=(pw*25.4)/72;
	ph=(ph*25.4)/72;
	ArgyllPaperSize ps(int(pw+0.5),int(ph+0.5));
	int res=core.FindInt("ChartResolution");
	Argyll_TestChart_TIFFs *tiffs=core.RenderTIFFs(fnroot,ps,inst,res,FindInt("ChartSeed"),p);
	delete tiffs;
	free(fnroot);
}


TIXFile *LinProject::GetRawChannelLinChart(const char *id)
{
	char *fnroot=GetTempRoot(id,".ti1");
	TIXFile *ti1=new TIXFile(fnroot);
	free(fnroot);
	return(ti1);
}


Argyll_TestChart_TIFFs *LinProject::GetRawChannelLinTIFFs(const char *id)
{
	char *fnroot=GetTempRoot(id);
	Argyll_TestChart_TIFFs *tiffs=new Argyll_TestChart_TIFFs(fnroot);
	free(fnroot);
	return(tiffs);
}


// These two functions differ only from the RawChannel variants by default ID.
TIXFile *LinProject::GetDropSizeLinChart(const char *id)
{
	return(GetRawChannelLinChart(id));
}


Argyll_TestChart_TIFFs *LinProject::GetDropSizeLinTIFFs(const char *id)
{
	return(GetRawChannelLinTIFFs(id));
}


const char *LinProject::GetChannelMode()
{
	return(stp_get_string_parameter(gprinter.stpvars,"InputImageType"));
}


void LinProject::SetChannelMode(const char *mode)
{
	stp_set_string_parameter(gprinter.stpvars,"InputImageType",mode);
}


void LinProject::SetProjectType(enum LinProjectType type)
{
	projecttype=type;
}


void LinProject::SetColorants(const char *colorants)
{
	switch(projecttype)
	{
		case LINPROJECTTYPE_LINEARIZATION:
			SetString("Colorants",colorants);
			break;
		case LINPROJECTTYPE_TUNING:
			SetString("DropsizeColorants",colorants);
			break;
		default:
			throw "Unknown project type!";
			break;
	}
}


const char *LinProject::GetColorants()
{
	switch(projecttype)
	{
		case LINPROJECTTYPE_LINEARIZATION:
			return(FindString("Colorants"));
			break;
		case LINPROJECTTYPE_TUNING:
			return(FindString("DropsizeColorants"));
			break;
		default:
			throw "Unknown project type!";
			break;
	}
}


int LinProject::GetStepsPerChannel()
{
	switch(projecttype)
	{
		case LINPROJECTTYPE_LINEARIZATION:
			return(FindInt("StepsPerChannel"));
			break;
		case LINPROJECTTYPE_TUNING:
			return(FindInt("DropSizeStepsPerChannel"));
			break;
		default:
			throw "Unknown project type!";
			break;
	}
}

void LinProject::SetStepsPerChannel(int steps)
{
	switch(projecttype)
	{
		case LINPROJECTTYPE_LINEARIZATION:
			SetInt("StepsPerChannel",steps);
			break;
		case LINPROJECTTYPE_TUNING:
			SetInt("DropSizeStepsPerChannel",steps);
			break;
		default:
			throw "Unknown project type!";
			break;
	}
}



void LinProject::Print(ImageSource *is,int xpos,int ypos,bool bypasslinearization)
{
	const char *mode=GetChannelMode();
	if(strcmp(mode,"RGB")==0)
	{
		is=new ImageSource_DeviceNToRGB(is);
	}
	else if(strcmp(mode,"CMYK")==0)
	{
		is=new ImageSource_DeviceNToCMYK(is);
	}
	cerr << "*** LinProject: Printing at position: " << xpos << ", " << ypos << endl;
	if(bypasslinearization)
		gprinter.GPrinter::Print(is,xpos,ypos);
	else
		gprinter.Print(is,xpos,ypos);
}

