#include <iostream>

#include <getopt.h>

#include <gtk/gtk.h>

#include "progresstext.h"
#include "progressbar.h"

#include "generaldialogs.h"

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_devicen_remap.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_greyscale.h"
#include "imagesource_colorize.h"

#include "testpatterns.h"

#include "lincore.h"
#include "linproject.h"
#include "linearizedprinter.h"
#include "channelresponse.h"

#include "config.h"


using namespace std;


class PrintColorizedCore : public LinearizationCore
{
	public:
	PrintColorizedCore() : LinearizationCore(), project(NULL), presetname(NULL)
	{

	}
	~PrintColorizedCore()
	{
		if(project)
			delete project;
	}
	void ParseOptions(int argc, char *argv[])
	{
		static struct option long_options[] =
		{
			{"help",no_argument,NULL,'h'},
			{"version",no_argument,NULL,'v'},
			{"preset",required_argument,NULL,'p'},
			{0, 0, 0, 0}
		};

		while(1)
		{
			int c;
			c = getopt_long(argc,argv,"hvp:",long_options,NULL);
			if(c==-1)
				break;
			switch (c)
			{
				case 'h':
					Usage(argv[0]);
					throw 0;
					break;
				case 'v':
					cerr << "PrintColorized - part of " << PACKAGE_STRING << endl;
					throw 0;
					break;
				case 'p':
					presetname=optarg;
					break;
			}
		}
	}
	void Usage(const char *name)
	{
		cerr << "Usage: " << name << " [options] Image1.tif [Image2.tif] ..." << endl;
		cerr << "\t -h --help\t\tdisplay this message" << endl;
		cerr << "\t -v --version\t\tdisplay version" << endl;
		cerr << "\t -p --preset\t\tfetch printer settings from a" << endl;
		cerr << "\t\t\t\tPhotoPrint preset or .gplin file" << endl;
	}
	void Load()
	{
		project=new LinProject(*this,presetname);
	}
	LinProject *project;
	protected:
	const char *presetname;
};


class PrintColorizedUI : public PrintColorizedCore
{
	public:
	PrintColorizedUI(int argc,char **argv) : PrintColorizedCore()
	{
		ParseOptions(argc,argv);
		if(!presetname)
		{
			char *fn=File_Dialog("Please select a GPLin preset file...",NULL);
			if(fn)
			{
				presetname=fn;
			}
			else
				throw "PrintColorized requires a GPLin preset file for printer settings.";
		}
		Load();

		if(argc>optind)
		{
			for(int file=optind;file<argc;++file)
			{
				try
				{
					Print(argv[file]);
				}
				catch(const char *err)
				{
					ErrorMessage_Dialog(err);
				}
			}
		}
		else
		{
			char *fn=File_Dialog("Please select image for printing...",NULL);
			if(fn)
			{
				try
				{
					Print(fn);
				}
				catch(const char *err)
				{
					ErrorMessage_Dialog(err);
				}
				free(fn);
			}
		}
	}
	~PrintColorizedUI()
	{
	}
	void Print(const char *file)
	{
		GPColorantList colorants(&project->gprinter);
		colorants.SetEnabledColorants(project->GetColorants());
		int colcount=colorants.GetColorantCount();

		// Create a colour which will be applied to the greyscale image
		ISDeviceNValue colour(colcount,0);
		for(int i=0;i<colcount;++i)
		{
			if(colorants[i]->GetEnabled())
				colour[i]=IS_SAMPLEMAX;
		}

		cerr << "Printing " << file << endl;

		ProgressBar p("Printing...",true);
		project->gprinter.SetProgress(&p);

		ImageSource *is=ISLoadImage(file);
		if(STRIP_ALPHA(is->type)!=IS_TYPE_GREY)
			is=new ImageSource_Greyscale(is);

		is=new ImageSource_Colorize(is,IS_TYPE_DEVICEN,colour);

		project->gprinter.Print(is,0,0);
		delete is;
	}
	protected:
};


int main(int argc, char **argv)
{
	gtk_init(&argc,&argv);
	try
	{
		PrintColorizedUI(argc,argv);

	}
	catch(const char *err)
	{
		ErrorMessage_Dialog(err);
	}
	catch(int)
	{
	}
	return(0);
}

