/*
 * stepselector.cpp
 * Provides a step-by-step list of tasks to be done
 * with an icon beside each one to indicate completed status.
 *
 * Copyright (c) 2008 by Alastair M. Robinson
 * Distributed under the terms of the GNU General Public License -
 * see the file named "COPYING" for more details.
 *
 */

#include <iostream>

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <gtk/gtkentry.h>
#include <gtk/gtklist.h>
#include <gtk/gtktreeselection.h>
#include <gtk/gtkscrolledwindow.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include <gutenprint/gutenprint.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "imagesource/imagesource_gdkpixbuf.h"
#include "imagesource/imagesource_greyscale.h"
#include "imagesource/imagesource_devicen_preview.h"
#include "imagesource/devicencolorant.h"
#include "imagesource/pixbuf_from_imagesource.h"

#include "miscwidgets/generaldialogs.h"
#include "gpdevicensupport.h"
#include "stepselector.h"

#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)


using namespace std;

enum {
	CHANGED_SIGNAL,
	LAST_SIGNAL
};

static guint stepselector_signals[LAST_SIGNAL] = { 0 };

static void stepselector_class_init (StepSelectorClass *klass);
static void stepselector_init (StepSelector *sel);


static void selection_changed(GtkTreeSelection *select,gpointer user_data)
{
	StepSelector *pe=STEPSELECTOR(user_data);
	g_signal_emit(G_OBJECT(pe),stepselector_signals[CHANGED_SIGNAL], 0);
};


// Declare multiple columns for GTK's tree model.  The columnns can contain widgets,
// not just static display, and don't have to be visible, so we can include userdata.
enum
{
  ICON_COLUMN=0,			// Pixbuf
  LABEL_COLUMN,			// Colorant name (string)
						// Now non-displayed fields:
  POINTER_COLUMN,		// Pointer to Step
  NUM_COLUMNS
};


static void populate_list(StepSelector *es)
{
	if(es->list)
	{
		GtkTreeIter iter;

		GdkPixbuf *icon=NULL;
		GdkPixdata pd;
//	GError *err;
//	if(!gdk_pixdata_deserialize(&pd,sizeof(my_pixbuf),my_pixbuf,&err))
//		throw(err->message);

//	if(!(icon=gdk_pixbuf_from_pixdata(&pd,false,&err)))
//		throw(err->message);

		StepSelector_Step *step=es->list->FirstStep();
		while(step)
		{
			cerr << "Processing step: " << step->GetUIName() << endl;
			gtk_tree_store_append (GTK_TREE_STORE(es->treestore), &iter, NULL);
			gtk_tree_store_set (GTK_TREE_STORE(es->treestore), &iter,
//				ACTIVE_COLUMN, col->GetEnabled(),
				LABEL_COLUMN, step->GetUIName(),
				ICON_COLUMN, NULL,
				POINTER_COLUMN, step,
				-1);
			step=step->NextStep();
	    }
//	g_object_unref(icon);
	}
}


static void setup_renderers(StepSelector *sel)
{
	gint col_offset;
	GtkCellRenderer *renderer;

	// Active column
//	renderer = gtk_cell_renderer_toggle_new ();

//	g_signal_connect (renderer, "toggled", G_CALLBACK (item_toggled), sel);

//	col_offset = gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (sel->treeview),
//		-1, _("Active"),
//		renderer,
//		"active",
//		ACTIVE_COLUMN,
//		NULL);

	// Pixbuf column
	renderer=gtk_cell_renderer_pixbuf_new();
	col_offset=gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW (sel->treeview),
		-1,_("Icon"),
		renderer,
		"pixbuf",
		ICON_COLUMN,
		NULL);

	// column for names
	renderer = gtk_cell_renderer_text_new ();
	g_object_set (renderer, "xalign", 0.0, NULL);
	g_object_set (renderer, "xpad", 10, NULL);
	
	col_offset = gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (sel->treeview),
	    -1, _("Name"),
	    renderer, "text",
	    LABEL_COLUMN,
	    NULL);
}


GtkWidget*
stepselector_new (StepSelector_StepList *list)
{
	StepSelector *c=STEPSELECTOR(g_object_new (stepselector_get_type (), NULL));

	c->list=list;

	c->treestore = gtk_tree_store_new (NUM_COLUMNS,
		GDK_TYPE_PIXBUF,
		G_TYPE_STRING,
		G_TYPE_POINTER
		);

	GtkWidget *sw = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw),GTK_SHADOW_ETCHED_IN);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),GTK_POLICY_NEVER,GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX (c), sw, TRUE, TRUE, 0);
	gtk_widget_show(sw);

	c->treeview=gtk_tree_view_new_with_model(GTK_TREE_MODEL(c->treestore));
	gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(c->treeview),FALSE);
	setup_renderers(c);

	GtkTreeSelection *select;
	select = gtk_tree_view_get_selection (GTK_TREE_VIEW (c->treeview));

	gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
	g_signal_connect (G_OBJECT (select), "changed",
		G_CALLBACK (selection_changed),c);

	gtk_container_add(GTK_CONTAINER(sw),c->treeview);
	gtk_widget_show(c->treeview);

	GtkWidget *hbox=gtk_hbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(c),hbox,FALSE,FALSE,0);
	gtk_widget_show(hbox);

	populate_list(c);
	
	return(GTK_WIDGET(c));
}


GType
stepselector_get_type (void)
{
	static GType stpuic_type = 0;

	if (!stpuic_type)
	{
		static const GTypeInfo stepselector_info =
		{
			sizeof (StepSelectorClass),
			NULL, /* base_init */
			NULL, /* base_finalize */
			(GClassInitFunc) stepselector_class_init,
			NULL, /* class_finalize */
			NULL, /* class_data */
			sizeof (StepSelector),
			0,
			(GInstanceInitFunc) stepselector_init,
		};
		stpuic_type = g_type_register_static (GTK_TYPE_VBOX, "StepSelector", &stepselector_info, GTypeFlags(0));
	}
	return stpuic_type;
}


static void *parent_class=NULL;

static void stepselector_destroy(GtkObject *object)
{
	if(object && IS_STEPSELECTOR(object))
	{
//		StepSelector *es=STEPSELECTOR(object);

		// FIXME - should free the colorant names here...

		if (GTK_OBJECT_CLASS (parent_class)->destroy)
			(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
	}
}


static void
stepselector_class_init (StepSelectorClass *cls)
{
	GtkObjectClass *object_class=(GtkObjectClass *)cls;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	object_class->destroy = stepselector_destroy;

	stepselector_signals[CHANGED_SIGNAL] =
	g_signal_new ("changed",
		G_TYPE_FROM_CLASS (cls),
		GSignalFlags(G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION),
		G_STRUCT_OFFSET (StepSelectorClass, changed),
		NULL, NULL,
		g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
}


static void
stepselector_init (StepSelector *c)
{
}


// Rebuilds the listview from the ColorantList
void stepselector_refresh(StepSelector *c)
{
	if(c)
	{
		// FIXME - should go through the list first, freeing
		// pixbufs and names?
		gtk_tree_store_clear(c->treestore);
		populate_list(c);
	}
}


void stepselector_set_steps(StepSelector *c,StepSelector_StepList *list)
{
	gtk_tree_store_clear(c->treestore);
	c->list=list;
	populate_list(c);
}


StepSelector_Step *stepselector_get(StepSelector *sel)
{
	StepSelector_Step *result=NULL;
	GtkTreeSelection *select = gtk_tree_view_get_selection (GTK_TREE_VIEW (sel->treeview));

	GtkTreeIter iter;
	GtkTreeModel *model;
	if(gtk_tree_selection_get_selected (select, &model, &iter))
	{
		gtk_tree_model_get (model, &iter, POINTER_COLUMN, &result, -1);
	}
	return(result);
}


int stepselector_get_index(StepSelector *sel)
{
	StepSelector_Step *result=stepselector_get(sel);
	if(result)
		return(result->GetIndex());
	else
		return(0);
}


void stepselector_set_index(StepSelector *sel,int idx)
{
	char path[]={'0','0'};
	if(idx>9)
	{
		path[0]+=idx/10;
		path[1]+=idx%10;
	}
	else
	{
		path[0]+=idx%10;
		path[1]=0;
	}
	GtkTreePath *tmppath=gtk_tree_path_new_from_string(path);
	gtk_tree_view_set_cursor(GTK_TREE_VIEW(sel->treeview),tmppath,NULL,FALSE);
	gtk_tree_view_row_activated(GTK_TREE_VIEW(sel->treeview),tmppath,NULL);
	gtk_tree_path_free(tmppath);
}


// StepSelect_Step classes

StepSelector_Step::StepSelector_Step(StepSelector_StepList &header,const char *uiname)
	: header(header), status(STEPSELECTOR_STATUS_NOTDONE), uiname(NULL), index(0), nextstep(NULL), prevstep(NULL)
{
	this->uiname=strdup(uiname);
	if((prevstep=header.firststep))
	{
		++index;
		while(prevstep->nextstep)
		{
			++index;
			prevstep=prevstep->nextstep;
		}
		prevstep->nextstep=this;
	}
	else
		header.firststep=this;	
}


StepSelector_Step::~StepSelector_Step()
{
	if(uiname)
		free(uiname);
	if(nextstep)
		nextstep->prevstep=prevstep;
	if(prevstep)
		prevstep->nextstep=nextstep;
	else
		header.firststep=nextstep;
}


const char *StepSelector_Step::GetUIName()
{
	return(gettext(uiname));
}


StepSelector_Step *StepSelector_Step::NextStep()
{
	return(nextstep);
}


StepSelector_Step *StepSelector_Step::PrevStep()
{
	return(prevstep);
}


enum StepSelector_Step_Status StepSelector_Step::GetStatus()
{
	return(status);
}


int StepSelector_Step::GetIndex()
{
	return(index);
}

// StepSelector_StepList header class

StepSelector_StepList::StepSelector_StepList() : firststep(NULL)
{
}

StepSelector_StepList::~StepSelector_StepList()
{
	while(firststep)
		delete firststep;
}

StepSelector_Step *StepSelector_StepList::FirstStep()
{
	return(firststep);
}

StepSelector_Step *StepSelector_StepList::operator[](int i)
{
	StepSelector_Step *step=FirstStep();
	while(step && i)
	{
		step=step->NextStep();
		--i;
	}
	return(step);
}

