#include <iostream>

#include "miscwidgets/generaldialogs.h"
#include "progressbar.h"
#include "support/util.h"
#include "stpui_widgets/stpui_combo.h"
#include "stpui_widgets/dimension.h"

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_devicen_remap.h"
#include "imagesource_dropsizesegment.h"

#include "tuning.h"
#include "miscwidgets/colorantselector.h"
#include "testpatterns.h"

#include "stepwidget_dropsizeinklimit.h"

#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)


// DropSize Ink Limit "step" - self contained class which builds and links all
// the widgets required.  Use the "widget" member to obtain a GtkWidget which can
// be packed into a window

// This is used to pick a suitable starting density for each dropsize - used to avoid paper saturation when
// printing the main dropsize calibration target

StepWidget_DropSizeInkLimit::StepWidget_DropSizeInkLimit(ProjectGUI *gui) : StepWidget(gui), colorants(&gui->gprinter)
{
	// Ensure the dither algorithm is set to "Dropsize Segmented"...
	stp_set_string_parameter(gui->gprinter.stpvars,"DitherAlgorithm","Segmented");
	// ... and the colour correction to "Raw"...
	stp_set_string_parameter(gui->gprinter.stpvars,"ColorCorrection","Raw");

	widget=gtk_vbox_new(FALSE,0);

	GtkWidget *hbox=gtk_hbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(widget),hbox,TRUE,TRUE,0);
	gtk_widget_show(hbox);


	// Colorantselector
	cerr << "Finding Colorants" << endl;
	colorants.SetEnabledColorants(gui->FindString("DropsizeColorants"));
	cerr << "Done" << endl;
	colorantsel=colorantselector_new(&colorants);
	gtk_box_pack_start(GTK_BOX(hbox),colorantsel,FALSE,FALSE,0);
	gtk_widget_show(colorantsel);
	
	gtk_widget_set_sensitive(GTK_WIDGET(colorantsel),FALSE);


	GtkWidget *vbox=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,8);
	gtk_widget_show(vbox);

	GtkWidget *table=gtk_table_new(3,2,FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table),8);

	gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
	gtk_widget_show(table);


	// Papersize selector

	GtkWidget *tmp=gtk_label_new(_("Paper size:"));
//	gtk_box_pack_start(GTK_BOX(vbox),tmp,FALSE,FALSE,0);
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,0,1);
	gtk_widget_show(tmp);

	papersize=stpui_combo_new(gui->gprinter.stpvars,"PageSize",NULL);
	gtk_table_attach_defaults(GTK_TABLE(table),papersize,1,3,0,1);
//	gtk_box_pack_start(GTK_BOX(vbox),papersize,FALSE,FALSE,4);
	gtk_widget_show(papersize);


	// Top edge selector

	tmp=gtk_label_new(_("Top edge:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,1,2);
	gtk_widget_show(tmp);

	topedge=dimension_new(0.0,1000.0,UNIT_MILLIMETERS);
	dimension_set_pt(DIMENSION(topedge),14);
	gtk_table_attach_defaults(GTK_TABLE(table),topedge,1,3,1,2);
	gtk_widget_show(topedge);


	// Starting density

	tmp=gtk_label_new(_("Base densities for dropsizes:"));
//	gtk_box_pack_start(GTK_BOX(vbox),tmp,FALSE,FALSE,0);
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,3,2,3);
	gtk_widget_show(tmp);

	tmp=gtk_label_new(_("Small:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,3,4);
	gtk_widget_show(tmp);

	tmp=gtk_label_new(_("Medium:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,1,2,3,4);
	gtk_widget_show(tmp);

	tmp=gtk_label_new(_("Large:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,2,3,3,4);
	gtk_widget_show(tmp);


	density[0]=gtk_spin_button_new_with_range(0,1.0,0.005);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(density[0]),gui->gprinter.tuning.FindFloat("DropSizeLimit0"));
	gtk_table_attach_defaults(GTK_TABLE(table),density[0],0,1,4,5);
	g_signal_connect(G_OBJECT(density[0]),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(density[0]);

	density[1]=gtk_spin_button_new_with_range(0,1.0,0.005);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(density[1]),gui->gprinter.tuning.FindFloat("DropSizeLimit1"));
	gtk_table_attach_defaults(GTK_TABLE(table),density[1],1,2,4,5);
	g_signal_connect(G_OBJECT(density[1]),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(density[1]);

	density[2]=gtk_spin_button_new_with_range(0,1.0,0.005);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(density[2]),gui->gprinter.tuning.FindFloat("DropSizeLimit2"));
	gtk_table_attach_defaults(GTK_TABLE(table),density[2],2,3,4,5);
	g_signal_connect(G_OBJECT(density[2]),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(density[2]);


	// Print button

	tmp=gtk_button_new_with_label(_("Print Density Evaluation Chart"));
//	gtk_box_pack_start(GTK_BOX(vbox),tmp,FALSE,FALSE,4);
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,3,5,6);
	gtk_widget_show(tmp);

	g_signal_connect(G_OBJECT(tmp),"clicked",G_CALLBACK(printclicked),this);


	// Spacer box
	tmp=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(vbox),tmp,TRUE,TRUE,0);
	gtk_widget_show(tmp);
}


// Text for the Directions box

const char *StepWidget_DropSizeInkLimit::GetDirections()
{
	return(_("Please use this step to determine a suitable base density for each "
		"dropsize.  These are specified in ascending order of drop size - so "
		"0.5, 0.25, 0.1 might be a suitable starting point.  If there's any "
		"saturation visible in any of the three sweeps, reduce the density and try "
		"again.  Similarly, increase the density for any sweep which fails to reach "
		"a reasonable degree of blackness.  (Note, depending on the chosen resolution "
		"the smaller dropsizes may not reach full blackness even at full density.)"));
}


void StepWidget_DropSizeInkLimit::SaveChanges()
{
	const char *ps=stp_get_string_parameter(gui->gprinter.stpvars,"PageSize");
	cerr << "Pagesize is currently set to: " << ps << endl;

	float t;
	t=gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[0]));
	gui->gprinter.tuning.SetFloat("DropSizeLimit0",t);
	t=gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[1]));
	gui->gprinter.tuning.SetFloat("DropSizeLimit1",t);
	t=gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[2]));
	gui->gprinter.tuning.SetFloat("DropSizeLimit2",t);

	gui->Save();

	modified=false;
}


void StepWidget_DropSizeInkLimit::Print()
{
	SaveChanges();

	// We render the Density Sweep pattern to about 4 inches wide.  Since the main
	// dropsize calibration chart has only a limited number of patches, we can use
	// one side of the paper for the density sweeps, and the other for the charts.
	ImageSource *is=MakeDensityEvalSweep(colorants,780,80,1,200);

	// Create a DeviceN Value here from the three dropsize base densities.
	ISDeviceNValue densities(is->samplesperpixel);
	densities[0]=IS_SAMPLEMAX*gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[0]));
	densities[1]=IS_SAMPLEMAX*gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[1]));
	densities[2]=IS_SAMPLEMAX*gtk_spin_button_get_value(GTK_SPIN_BUTTON(density[2]));

	is=new ImageSource_ScaleDensity(is,densities);

	// Now convert to drop-size-segnmented...
	is=new ImageSource_DropSizeSegment(is);

	// Remap to printer DeviceN
	GPColorantList gpcolorants(&gui->gprinter);
	int colcount=gpcolorants.GetColorantCount();
	int blackidx=gpcolorants.GetColorantIndex("Black");
	if(blackidx<0)
		blackidx=gpcolorants.GetColorantIndex("Photo Black");
	cerr << "Got black index:" << blackidx << endl;
	int maptable[2]={0,0};
	maptable[0]=blackidx;
	int magidx=gpcolorants.GetColorantIndex("Magenta");
	if(magidx<0)
		magidx=gpcolorants.GetColorantIndex("Vivid Magenta");
	cerr << "Got magenta index:" << magidx << endl;
	maptable[1]=magidx;

	ImageSource_DeviceN_Remap dst(is,maptable,colcount);

	float top=dimension_get_pt(DIMENSION(topedge));

	gui->Print(&dst,20,top,true);	// Bypass linearization since we're applying density manually.
	SaveChanges();
}


// Callback function to mark the step as modified...
void StepWidget_DropSizeInkLimit::markmodified(GtkWidget *wid,gpointer data)
{
	StepWidget_DropSizeInkLimit *step=(StepWidget_DropSizeInkLimit *)data;
	step->modified=true;
}


// Callback function for the Print button
void StepWidget_DropSizeInkLimit::printclicked(GtkWidget *wid,gpointer data)
{
	StepWidget_DropSizeInkLimit *step=(StepWidget_DropSizeInkLimit *)data;
	try
	{
		step->Print();
	}
	catch(const char *err)
	{
		ErrorMessage_Dialog(err,step->gui->window);
	}
}

