#include <iostream>

#include <cairo.h>

#include "miscwidgets/generaldialogs.h"
#include "progressbar.h"
#include "support/util.h"
#include "stpui_widgets/stpui_combo.h"
#include "stpui_widgets/dimension.h"

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_modifiedgamma.h"
#include "imagesource/imagesource_gamma.h"
#include "imagesource/imagesource_devicen_remap.h"

#include "testpatterns.h"
#include "miscwidgets/colorantselector.h"
#include "dropsize.h"

#include "stepwidget_dropsizeratios.h"

#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)


// Class containing data needed by the graph-drawing callback functions

class DropSizeRatios_Payload
{
	public:
	DropSizeRatios_Payload(StepWidget_DropSizeRatios &step)
		: step(step), ti3(NULL), smalldrops(NULL), mediumdrops(NULL), largedrops(NULL)
	{
		ti3=step.chans;
		gamma=step.gui->gprinter.tuning.FindFloat("TuningGamma");
		ChannelResponse *c=ti3->GetChannel("Light Light Black");
		if(c)
			smalldrops=new DropSizeResponse(c,step.gui->gprinter.tuning.FindFloat("DropSizeLimit0"),gamma);
		else
			throw "Couldn't get channel: Light Light Black";

		c=ti3->GetChannel("Light Black");
		if(c)
			mediumdrops=new DropSizeResponse(c,step.gui->gprinter.tuning.FindFloat("DropSizeLimit1"),gamma);
		else
			throw "Couldn't get channel: Light Black";

		c=ti3->GetChannel("Black");
		if(c)
			largedrops=new DropSizeResponse(c,step.gui->gprinter.tuning.FindFloat("DropSizeLimit2"),gamma);
		else
			throw "Couldn't get channel: Black";
	}
	~DropSizeRatios_Payload()
	{
		if(smalldrops)
			delete smalldrops;
		if(mediumdrops)
			delete mediumdrops;
		if(largedrops)
			delete largedrops;
	}
	DropSizeRatios FindDropSizeRatios(double input)
	{
		double L=smalldrops->InputToL(input);
		double gp1=smalldrops->InputToGP(input);
		double gp2=mediumdrops->LToGP(L);
		double gp3=largedrops->LToGP(L);

		double L2=mediumdrops->InputToL(input);
		double gp4=mediumdrops->InputToGP(input);
		double gp5=largedrops->LToGP(L2);

		DropSizeRatios result(gp2/gp1,gp3/gp1,gp5/gp4);
		return(result);
	}
	void UpdateRatios()
	{
		double input=x;
		input/=width;
		if(input<.01)
			input=.01;
		if(input>1.0)
			input=1.0;

		double lininput=100.0*(1.0-ImageSource_ModifiedGamma::InverseModifiedGamma(1.0-input,gamma));
		ratios=FindDropSizeRatios(lininput);

		const char *fmt="Small dropsize density: %0.3lf   -   medium/small: %0.3lf   -   large/small: %0.3lf   -   large/medium: %0.3lf";
		char *text=(char *)malloc(strlen(fmt));
		sprintf(text,fmt,smalldrops->InputToGP(lininput),ratios.ratio1,ratios.ratio2,ratios.ratio3);
		gtk_label_set_text(GTK_LABEL(step.status),text);
		free(text);

//		Double-check the values here...:
		double smallinput=smalldrops->InputToGP(lininput);
		double l_small=smalldrops->InputToL(smalldrops->GPToInput(smallinput));
		cerr << "in_small: " << smallinput << endl;
		cerr << "L_small: " << l_small << endl;

		double medinput=smallinput*ratios.ratio1;
		double l_medium=mediumdrops->InputToL(mediumdrops->GPToInput(medinput));
		cerr << "in_medium: " << medinput << endl;
		cerr << "L_medium: " << l_medium << endl;

		double largeinput=smallinput*ratios.ratio2;
		double l_large=largedrops->InputToL(largedrops->GPToInput(largeinput));
		cerr << "in_large: " << largeinput << endl;
		cerr << "L_large: " << l_large << endl;

		double largeinput2=medinput*ratios.ratio3;
		double l_large2=largedrops->InputToL(largedrops->GPToInput(largeinput2));
		cerr << "in_large2: " << largeinput2 << endl;
		cerr << "L_large2: " << l_large2 << endl << endl;

//		tell the canvas widget that it needs to redraw itself
//		gtk_widget_queue_draw (widget);
	}
	static gboolean event_press(GtkWidget *widget, GdkEventButton *bev, DropSizeRatios_Payload *payload)
	{
		payload->x=bev->x;
		payload->y=bev->y;
		payload->UpdateRatios();
		return TRUE;
	}
	static gboolean event_motion(GtkWidget *widget, GdkEventMotion *mev, DropSizeRatios_Payload *payload)
	{
		payload->x=mev->x;
		payload->y=mev->y;
		payload->UpdateRatios();
		return TRUE;
	}
	// Graph drawing functions
	// Input vs. DropSize Ratios
	static void paint_input_vs_subchannel_ratio(GtkWidget *widget,GdkEventExpose *eev,DropSizeRatios_Payload *payload)
	{
		int width = payload->width  = widget->allocation.width;
		int height = payload->height = widget->allocation.height;

		cairo_t *cr = gdk_cairo_create (widget->window);
	 
		cairo_set_source_rgb (cr, 1,1,1);
		cairo_paint (cr);

		cairo_set_source_rgb (cr, 0,0,0);
		cairo_set_line_width(cr,1.0);
		cairo_move_to (cr, 1,0);
		cairo_line_to(cr,1,height);
		cairo_move_to(cr,0,height-1);
		cairo_line_to (cr, width,height-1);
		cairo_stroke (cr);

	    cairo_select_font_face (cr, "Sans", CAIRO_FONT_SLANT_ITALIC,
	                                        CAIRO_FONT_WEIGHT_BOLD);
		cairo_set_font_size (cr, 9);
		cairo_move_to (cr, width-75, height-3);
		cairo_show_text (cr, "Input 100%");

		cairo_set_source_rgb (cr, 0,0,0);

		bool first=true;

		for(int i=1;i<=100;++i)
		{
			float input=i;
			DropSizeRatios ratios=payload->FindDropSizeRatios(input);
			input=1.0-ImageSource_ModifiedGamma::ModifiedGamma(1.0-(input/100.0),payload->gamma);

			if(first)
				cairo_move_to(cr,(input*width),height-(ratios.ratio1*height));
			else
				cairo_line_to(cr,(input*width),height-(ratios.ratio1*height));
			cairo_rel_line_to(cr,2,2);
			cairo_rel_move_to(cr,0,-2);
			cairo_rel_line_to(cr,-2,2);
			cairo_move_to(cr,(input*width),height-(ratios.ratio1*height));
			first=false;
		}
		cairo_stroke(cr);

		first=true;
		for(int i=1;i<=100;++i)
		{
			float input=i;
			DropSizeRatios ratios=payload->FindDropSizeRatios(input);
			input=1.0-ImageSource_ModifiedGamma::ModifiedGamma(1.0-(input/100.0),payload->gamma);

			if(first)
				cairo_move_to(cr,(input*width),height-(ratios.ratio3*height));
			else
				cairo_line_to(cr,(input*width),height-(ratios.ratio3*height));
			cairo_rel_line_to(cr,2,2);
			cairo_rel_move_to(cr,0,-2);
			cairo_rel_line_to(cr,-2,2);
			cairo_move_to(cr,(input*width),height-(ratios.ratio3*height));
			first=false;
		}
		cairo_stroke(cr);

		cairo_set_source_rgb (cr,0.5,0.5,0.5);

		first=true;
		for(int i=1;i<=100;++i)
		{
			float input=i;
			DropSizeRatios ratios=payload->FindDropSizeRatios(input);
			input=1.0-ImageSource_ModifiedGamma::ModifiedGamma(1.0-(input/100.0),payload->gamma);

			if(first)
				cairo_move_to(cr,(input*width),height-(ratios.ratio2*height));
			else
				cairo_line_to(cr,(input*width),height-(ratios.ratio2*height));
			cairo_rel_line_to(cr,2,2);
			cairo_rel_move_to(cr,0,-2);
			cairo_rel_line_to(cr,-2,2);
			cairo_move_to(cr,(input*width),height-(ratios.ratio2*height));
			first=false;
		}
		cairo_stroke(cr);

		cairo_destroy (cr);
	}
	protected:
	int x,y;
	int width,height;
	float gamma;
	StepWidget_DropSizeRatios &step;
	RawChannelResponses *ti3;
	DropSizeResponse *smalldrops;
	DropSizeResponse *mediumdrops;
	DropSizeResponse *largedrops;
	DropSizeRatios ratios;
};


// Individual channel ink limit "step" - self contained class which builds and links all
// the widgets required.  Use the "widget" member to obtain a GtkWidget which can
// be packed into a window

StepWidget_DropSizeRatios::StepWidget_DropSizeRatios(ProjectGUI *gui) : StepWidget(gui), chans(NULL), payload(NULL)
{
	char *fn=gui->GetTempRoot("DropSize",".ti3");
	try
	{
		chans=new RawChannelResponses(fn,gui->gprinter.linearization.FindFloat("BaseGamma"));
	}
	catch (const char *err)
	{
		cerr << "Couldn't get ChannelResponses - chart probably hasn't been read yet!" << endl;
	}
	free(fn);

	widget=gtk_vbox_new(FALSE,0);

	if(!chans)
		return;		// Bail out if there's no .ti3 file yet.

	payload=new DropSizeRatios_Payload(*this);

	notebook=gtk_notebook_new();
	gtk_box_pack_start(GTK_BOX(widget),notebook,TRUE,TRUE,0);
	gtk_widget_show(GTK_WIDGET(notebook));


//	Subchannel ratio graph

	GtkWidget *label=gtk_label_new(_("Input vs Dropsize Ratios"));
	gtk_widget_show(label);

	GtkWidget *canvas = gtk_drawing_area_new();
	gtk_widget_set_size_request(canvas, 50,50);
	g_signal_connect(G_OBJECT (canvas), "expose-event",
		G_CALLBACK (DropSizeRatios_Payload::paint_input_vs_subchannel_ratio),payload);

	gtk_widget_add_events (canvas,GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);

	g_signal_connect(G_OBJECT (canvas), "button_press_event",G_CALLBACK (DropSizeRatios_Payload::event_press),payload);
	g_signal_connect(G_OBJECT (canvas), "button_release_event",G_CALLBACK (DropSizeRatios_Payload::event_press),payload);
	g_signal_connect(G_OBJECT (canvas), "motion_notify_event",G_CALLBACK (DropSizeRatios_Payload::event_motion),payload);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook),GTK_WIDGET(canvas),GTK_WIDGET(label));
	gtk_widget_show(canvas);


//  Ratios

	status=gtk_label_new("");
	gtk_widget_show(status);
	gtk_box_pack_start(GTK_BOX(widget),status,FALSE,FALSE,0);

	payload->UpdateRatios();
}


StepWidget_DropSizeRatios::~StepWidget_DropSizeRatios()
{
	if(chans)
		delete chans;
	if(payload)
		delete payload;
}


// Text for the Directions box

const char *StepWidget_DropSizeRatios::GetDirections()
{
	if(chans)
	{
		return(_("This page is used to determine ratios between dropsizes.\n"));
	}
	else
	{
		return(_("Unable to open the DropSize .ti3 file, which will be created when reading the drop size calibration chart.\n\n"
				 "One this chart has been read, this page can be used to determine ratios between dropsizes.\n"));
	}
}


void StepWidget_DropSizeRatios::SaveChanges()
{
	gui->Save();

	modified=false;
}


// Callback function to mark the step as modified...
void StepWidget_DropSizeRatios::markmodified(GtkWidget *wid,gpointer data)
{
	StepWidget_DropSizeRatios *step=(StepWidget_DropSizeRatios *)data;
	step->modified=true;
}


