#include <iostream>

#include "miscwidgets/generaldialogs.h"
#include "progressbar.h"
#include "imageutils/jpegsaver.h"
#include "support/util.h"
#include "stpui_widgets/stpui_combo.h"

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_devicen_remap.h"

#include "argyll_uisupport.h"

#include "miscwidgets/colorantselector.h"
#include "stepwidget_generatecharts.h"

#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)


// Chart generation "step" - self contained class which builds and links all
// the widgets required.  Use the "widget" member to obtain a GtkWidget which can
// be packed into a window


StepWidget_GenerateChart::StepWidget_GenerateChart(ProjectGUI *gui) : StepWidget(gui), colorants(&gui->gprinter)
{
	widget=gtk_vbox_new(FALSE,0);
	GtkWidget *tmp;

	GtkWidget *hbox=gtk_hbox_new(FALSE,8);
	gtk_box_pack_start(GTK_BOX(widget),hbox,TRUE,TRUE,0);
	gtk_widget_show(hbox);
	
	// Colorant selector

	cerr << "Finding Colorants" << endl;
	colorants.SetEnabledColorants(gui->GetColorants());
	cerr << "Done" << endl;
	colorantsel=colorantselector_new(&colorants);
	g_signal_connect(G_OBJECT(colorantsel),"changed",G_CALLBACK(markmodified),this);
	gtk_box_pack_start(GTK_BOX(hbox),colorantsel,FALSE,FALSE,0);
	gtk_widget_show(colorantsel);

	// VBox for right-hand half of pane

	GtkWidget *vbox=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,0);
	gtk_widget_show(vbox);

	GtkWidget *table=gtk_table_new(3,2,FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table),8);

	gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
	gtk_widget_show(table);

	// Spacer box
	tmp=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(vbox),tmp,TRUE,TRUE,0);
	gtk_widget_show(tmp);

	// Instrument selector

	GtkWidget *label=gtk_label_new(_("Instrument:"));
	gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,0,1);
	gtk_widget_show(label);

	SimpleComboOptions comboopts;
	ArgyllUISupport::AddInstrumentComboOpts(comboopts);

	instrument=simplecombo_new(comboopts);
	cerr << "Finding Default Instrument" << endl;
	simplecombo_set(SIMPLECOMBO(instrument),gui->FindString("Instrument"));
	cerr << "Done" << endl;
	gtk_table_attach_defaults(GTK_TABLE(table),instrument,1,2,0,1);
	g_signal_connect(G_OBJECT(instrument),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(instrument);

	// Papersize selector
	
	label=gtk_label_new(_("Paper size:"));
	gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,1,2);
	gtk_widget_show(label);

	papersize=stpui_combo_new(gui->gprinter.stpvars,"PageSize",label);
	gtk_table_attach_defaults(GTK_TABLE(table),papersize,1,2,1,2);
	gtk_widget_show(papersize);

	// FIXME - add custom width and height widgets here.

	// Number of steps per channel

	label=gtk_label_new(_("Steps per channel:"));
	gtk_table_attach_defaults(GTK_TABLE(table),label,0,1,2,3);
	gtk_widget_show(label);

	stepsperchannel=gtk_spin_button_new_with_range(16.0,256.0,1.0);
	gtk_table_attach_defaults(GTK_TABLE(table),stepsperchannel,1,2,2,3);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(stepsperchannel),gui->GetStepsPerChannel());
	g_signal_connect(G_OBJECT(stepsperchannel),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(stepsperchannel);

	// FIXME - add an expander here with advanced options - including random seed.
}


const char *StepWidget_GenerateChart::GetDirections()
{
	return(_("Test-chart generation:  please choose a suitable papersize, and "
			"the type of instrument you'll be using to read the test chart(s).\n\n"
			"You can also select which channels you'd like to test, and choose how "
			"man steps per channel the generated testchart(s) will contain."));
}


void StepWidget_GenerateChart::SaveChanges()
{
	const char *ps=stp_get_string_parameter(gui->gprinter.stpvars,"PageSize");
	cerr << "Pagesize is currently set to: " << ps << endl;
	char *cols=colorants.GetEnabledColorants();
	cerr << "Selected colorants: " << cols << endl;
	gui->SetColorants(cols);
	free(cols);
	gui->SetString("Instrument",simplecombo_get(SIMPLECOMBO(instrument)));
	int spc=gtk_spin_button_get_value(GTK_SPIN_BUTTON(stepsperchannel));
	cerr << "Steps per channel" << spc << endl;
	gui->SetStepsPerChannel(spc);
	gui->Save();
	if(gui->GetFilename())
	{
		CreateChart();
		CreateThumbnails();
	}
	else
	{
		ErrorMessage_Dialog(_("Please choose a filename for the project before generating charts."));
	}
	modified=false;
}


void StepWidget_GenerateChart::CreateChart()
{
	ProgressBar p(_("Generating test target..."),true,gui->window);

	// We don't need to clean up TIFFs from a previous run after all,
	// The TestChart_Generator takes care of that for us.
	gui->CreateRawChannelLinChart("RawLin",&p);
}


void StepWidget_GenerateChart::CreateThumbnails()
{
	ProgressBar p(_("Generating Thumbnails..."),true,gui->window);

	// FIXME - Need to clean up thumbnails from a previous run
	TIXFile *ti1=gui->GetRawChannelLinChart("RawLin");
	if(ti1)
	{
		Argyll_TestChart_TIFFs *tiffs=gui->GetRawChannelLinTIFFs("RawLin");
		if(tiffs)
		{
			GPColorantList colorants(&gui->gprinter);
			int colcount=colorants.GetColorantCount();
			int *channelmap=ti1->CreateRawChannelMap(colorants);

			char *fn=tiffs->GetFirstFilename();
			while(fn)
			{
				ImageSource *is=ISLoadImage(fn);

				is=new ImageSource_ToDeviceN(is);
				if(channelmap)
					is=new ImageSource_DeviceN_Remap(is,channelmap,colcount);
				is=new ImageSource_DeviceN_Preview(is,&colorants);

				int w=128;
				int h=(w*is->height)/is->width;
				if(h>128)
				{
					h=128;
					w=(h*is->width)/is->height;
				}

				is=ISScaleImageBySize(is,w,h);

				// Save here

				char *fn2=BuildFilename(fn,"","jpg");

				JPEGSaver j(fn2,ImageSource_rp(is));
				j.SetProgress(&p);
				j.Save();

				free(fn2);
				free(fn);
				fn=tiffs->GetNextFilename();
			}

			delete[] channelmap;
		}
	}
}


// Callback function to mark the step as modified...
void StepWidget_GenerateChart::markmodified(GtkWidget *wid,gpointer data)
{
	StepWidget_GenerateChart *step=(StepWidget_GenerateChart *)data;
	step->modified=true;
}

