#include <iostream>

#include "miscwidgets/generaldialogs.h"
#include "progressbar.h"
#include "support/util.h"
#include "stpui_widgets/stpui_combo.h"
#include "stpui_widgets/dimension.h"

#include "imagesource/imagesource_util.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_modifiedgamma.h"
#include "imagesource/imagesource_gamma.h"
#include "imagesource/imagesource_devicen_remap.h"

#include "testpatterns.h"
#include "miscwidgets/colorantselector.h"

#include "stepwidget_totalinklimit.h"

#include "config.h"
#include "gettext.h"
#define _(x) gettext(x)
#define N_(x) gettext_noop(x)


// Total ink limit "step" - self contained class which builds and links all
// the widgets required.  Use the "widget" member to obtain a GtkWidget which can
// be packed into a window


StepWidget_TotalInkLimit::StepWidget_TotalInkLimit(ProjectGUI *gui) : StepWidget(gui), colorants(&gui->gprinter)
{
	widget=gtk_vbox_new(FALSE,0);

	GtkWidget *hbox=gtk_hbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(widget),hbox,TRUE,TRUE,0);
	gtk_widget_show(hbox);


	// Colorantselector
	cerr << "Finding Colorants" << endl;
	colorants.SetEnabledColorants(gui->GetColorants());
	cerr << "Done" << endl;
	colorantsel=colorantselector_new(&colorants);
	gtk_box_pack_start(GTK_BOX(hbox),colorantsel,FALSE,FALSE,0);
	gtk_widget_set_sensitive(colorantsel,false);
	gtk_widget_show(colorantsel);
	

	GtkWidget *vbox=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(hbox),vbox,TRUE,TRUE,8);
	gtk_widget_show(vbox);


	GtkWidget *table=gtk_table_new(3,3,FALSE);
	gtk_table_set_row_spacings(GTK_TABLE(table),8);

	gtk_box_pack_start(GTK_BOX(vbox),table,FALSE,FALSE,0);
	gtk_widget_show(table);


	// Spacer box
	GtkWidget *tmp=gtk_vbox_new(FALSE,0);
	gtk_box_pack_start(GTK_BOX(vbox),tmp,TRUE,TRUE,0);
	gtk_widget_show(tmp);


	// Papersize selector
	
	tmp=gtk_label_new(_("Paper size:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,0,1);
	gtk_widget_show(tmp);

	papersize=stpui_combo_new(gui->gprinter.stpvars,"PageSize",NULL);
	gtk_table_attach_defaults(GTK_TABLE(table),papersize,1,2,0,1);
	gtk_widget_show(papersize);


	// Top edge selector

	tmp=gtk_label_new(_("Top edge:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,1,2);
	gtk_widget_show(tmp);

	topedge=dimension_new(0.0,1000.0,UNIT_MILLIMETERS);
	dimension_set_pt(DIMENSION(topedge),14);
	gtk_table_attach_defaults(GTK_TABLE(table),topedge,1,2,1,2);
	gtk_widget_show(topedge);


	// Max inklimit for sweep

	tmp=gtk_label_new(_("Max ink limit:"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,2,3);
	gtk_widget_show(tmp);

	maxinklimit=gtk_spin_button_new_with_range(0,400,5);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(maxinklimit),gui->gprinter.linearization.FindFloat("TotalInkLimit"));
	g_signal_connect(G_OBJECT(maxinklimit),"changed",G_CALLBACK(markmodified),this);
	gtk_table_attach_defaults(GTK_TABLE(table),maxinklimit,1,2,2,3);
	gtk_widget_show(maxinklimit);


	// Print button

	tmp=gtk_button_new_with_label(_("Print Ink Limit Evaluation Chart"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,2,3,4);
	gtk_widget_show(tmp);

	g_signal_connect(G_OBJECT(tmp),"clicked",G_CALLBACK(printclicked),this);

	// Corrected Gamma

	tmp=gtk_label_new(_("Correction value\n(read from print):"));
	gtk_table_attach_defaults(GTK_TABLE(table),tmp,0,1,4,5);
	gtk_widget_show(tmp);

	newinklimit=gtk_spin_button_new_with_range(0,1.0,.05);
	gtk_table_attach_defaults(GTK_TABLE(table),newinklimit,1,2,4,5);
	float corrg=1.0;
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(newinklimit),corrg);
	g_signal_connect(G_OBJECT(newinklimit),"changed",G_CALLBACK(markmodified),this);
	gtk_widget_show(newinklimit);
}


// Text for the Directions box

const char *StepWidget_TotalInkLimit::GetDirections()
{
	return(_("Please use this page to determine a suitable "
		"total ink limit.  The printed sweep will consist of 100% black, and increasing coverage "
		"of the other colours, left to right, up to the maximum specified in 'Max ink limit'. "
		"Read off from the printed sweep the highest point with no paper saturation."));
}


void StepWidget_TotalInkLimit::SaveChanges()
{
	const char *ps=stp_get_string_parameter(gui->gprinter.stpvars,"PageSize");
	cerr << "Pagesize is currently set to: " << ps << endl;
	float initg=gtk_spin_button_get_value(GTK_SPIN_BUTTON(maxinklimit));
	float corrg=gtk_spin_button_get_value(GTK_SPIN_BUTTON(newinklimit));

	float corrx=100+(initg-100)*corrg;
	cerr << "Setting total ink limit to:" << corrx << endl;
	gui->gprinter.linearization.SetFloat("TotalInkLimit",corrx);

	char *cols=colorants.GetEnabledColorants();
	cerr << "Selected colorants: " << cols << endl;
	gui->SetColorants(cols);
	free(cols);

	gui->Save();

	modified=false;
}


void StepWidget_TotalInkLimit::Print()
{
	SaveChanges();
	float initl=gtk_spin_button_get_value(GTK_SPIN_BUTTON(maxinklimit));

	// We render the Gamma Sweep pattern to be a little under 8 inches wide
	ImageSource *is=MakeInkLimitEvalSweep(colorants,1560,240,initl,200);

	float top=dimension_get_pt(DIMENSION(topedge));
	cerr << "Top edge: " << top << endl;

	gui->Print(is,20,top,false);	// We don't want to bypass the linearization here
									// since by this stage we should have individual
									// channel ink limits and linearization curves.
	delete is;
}


// Callback function to mark the step as modified...
void StepWidget_TotalInkLimit::markmodified(GtkWidget *wid,gpointer data)
{
	StepWidget_TotalInkLimit *step=(StepWidget_TotalInkLimit *)data;
	step->modified=true;
}


// Callback function for the button
void StepWidget_TotalInkLimit::printclicked(GtkWidget *wid,gpointer data)
{
	StepWidget_TotalInkLimit *step=(StepWidget_TotalInkLimit *)data;
	try
	{
		step->Print();
	}
	catch(const char *err)
	{
		ErrorMessage_Dialog(err,step->gui->window);
	}
}

