
#include "imagesource/imagesource_montage.h"
#include "imagesource/imagesource_scaledensity.h"
#include "imagesource/imagesource_chequerboard.h"
#include "imagesource/imagesource_hsweep.h"
#include "imagesource/imagesource_hticks.h"

#include "gpdevicensupport.h"
#include "testpatterns.h"


// Render a sweep pattern for approximate gamma determination.
// This doesn't need to be accurate (which is just as well, because it isn't!)
// - just good enough to give us an approximate gamma so the test point when
// printing the linearization charts are reasonably well-distributed.

ImageSource *MakeGammaEvalSweep(GPColorantList &colorants,int width,int height,int res)
{
	int cols=colorants.GetColorantCount();
	int stripeheight=height/3;

	// We search for black, and throw an error if not found...
	int blk=-1;
	blk=colorants.GetColorantIndex("Black");
	if(blk==-1)
		blk=colorants.GetColorantIndex("Matte Black");
	if(blk==-1)
		throw "Error: Can't find black ink!";

	ISDeviceNValue bg(cols,0);	// Background colour
	ISDeviceNValue fg(cols,0);	// Foreground for ticks.
	fg[blk]=IS_SAMPLEMAX;	// Render ticks in black

	ImageSource_Montage *mon=new ImageSource_Montage(IS_TYPE_DEVICEN,res,cols);
	ImageSource_HTicks *ticks=new ImageSource_HTicks(width,stripeheight,IS_TYPE_DEVICEN,cols);
	ticks->SetBG(bg);
	ticks->SetFG(fg);
	mon->Add(ticks,0,0);	// Now owned by the montage

	ImageSource_HSweep *sweep = new ImageSource_HSweep(width,stripeheight,bg,fg,IS_TYPE_DEVICEN);
	mon->Add(sweep,0,stripeheight);	// Now owned by the montage
	
	ImageSource_Chequerboard *cb = new ImageSource_Chequerboard(width,stripeheight,IS_TYPE_DEVICEN,cols,4);
	cb->SetBG(bg);
	cb->SetFG(fg);
	mon->Add(cb,0,stripeheight*2);	// Now owned by the montage
	return(mon);
}


// Render a density evaluation pattern - this pattern should be printed at reduced density in
// Raw mode to determine absolute maximum ink levels for individual ink channels.

ImageSource *MakeDensityEvalSweep(GPColorantList &colorants,int width,int height,int columns,int res)
{
	int cols=colorants.GetColorantCount();
	int stripeheight=(2*height)/(cols+1);	// We'll render each pass as stripeheight pixels
											// and the row of "ticks" at half that height

	// We search for black, and throw an error if not found...
	int blk=-1;
	blk=colorants.GetColorantIndex("Black");
	if(blk==-1)
		blk=colorants.GetColorantIndex("Matte Black");
	if(blk==-1)
		throw "Error: Can't find black ink!";

	ISDeviceNValue bg(cols,0);	// Background colour
	ISDeviceNValue tfg(cols,0);	// Foreground for ticks.
	tfg[blk]=IS_SAMPLEMAX;	// Render ticks in black

	ImageSource_Montage *mon=new ImageSource_Montage(IS_TYPE_DEVICEN,res,cols);
	ImageSource_HTicks *ticks;

	for (int clm=0;clm<columns;++clm)
	{
		ticks=new ImageSource_HTicks(width/columns,stripeheight/2,IS_TYPE_DEVICEN,cols);
		ticks->SetBG(bg);
		ticks->SetFG(tfg);
		mon->Add(ticks,clm*(width/columns),0);	// Now owned by the montage
	}

	int slot=0;
	for(int i=0;i<cols;++i)
	{
		if((colorants[i])->GetEnabled())
		{
			cerr << "Generating sweep for colorant " << i << endl;
			ISDeviceNValue sfg(cols,0);
			sfg[i]=IS_SAMPLEMAX;
			int row=slot/columns;
			int clm=slot%columns;
			ImageSource *is=new ImageSource_HSweep(width/columns,stripeheight,bg,sfg,IS_TYPE_DEVICEN);
			mon->Add(is,(width/columns)*clm,stripeheight*row+stripeheight/2);
			++slot;
		}
	}
	return(mon);
}


// This routine constructs a sweep pattern for evaluating the ultimate ink limit.
// Black will be printed solidly, and the other colorants will be swept from zero to
// (inklimit-100) / (colorants-1)

ImageSource *MakeInkLimitEvalSweep(GPColorantList &colorants,int width,int height,int inklimit,int res)
{
	int cols=colorants.GetColorantCount();
	int stripeheight=height/3;	// We render the ticks at half the height of the sweep

	// We search for black, and throw an error if not found...
	int blk=-1;
	blk=colorants.GetColorantIndex("Black");
	if(blk==-1)
		blk=colorants.GetColorantIndex("Matte Black");
	if(blk==-1)
		throw "Error: Can't find black ink!";

	ISDeviceNValue bg(cols,0);	// Background colour
	ISDeviceNValue tfg(cols,0);	// Foreground for ticks.
	tfg[blk]=IS_SAMPLEMAX;	// Render ticks in black

	ImageSource_Montage *mon=new ImageSource_Montage(IS_TYPE_DEVICEN,res,cols);
	ImageSource_HTicks *ticks=new ImageSource_HTicks(width,stripeheight/2,IS_TYPE_DEVICEN,cols);
	ticks->SetBG(bg);
	ticks->SetFG(tfg);
	mon->Add(ticks,0,0);	// Now owned by the montage

	// Now we figure out the ink limit for remaining colorants.  We print black at 100%
	int activecols=0;

	for(int i=0;i<cols;++i)
	{
		if((colorants[i])->GetEnabled())
			++activecols;
	}
	cerr << "We have " << activecols << " active colorants" << endl;

	inklimit=(inklimit-100) / (activecols-1);

	cerr << "Ink limit for remaining channels: " << inklimit << endl;

	ISDeviceNValue sfg(cols,0);
	for(int i=0;i<cols;++i)
	{
		if((colorants[i])->GetEnabled())
		{
			if(i==blk)
				sfg[i]=IS_SAMPLEMAX;
			else
				sfg[i]=(IS_SAMPLEMAX*inklimit)/100;
			cerr << "Set channel " << i << " to " << sfg[i] << endl;
		}
	}
	bg[blk]=IS_SAMPLEMAX;

	ImageSource *is=new ImageSource_HSweep(width,stripeheight,bg,sfg,IS_TYPE_DEVICEN);
	mon->Add(is,0,stripeheight/2);

	return(mon);
}


