#include <iostream>
#include <string.h>

#include "config.h"
#include "support/thread.h"
#include "support/util.h"
#include "argyll_support.h"
#include "ticgats.h"

using namespace std;


CGATSFile::CGATSFile(const char *filename,const char *type) : cgatsfile(NULL)
{
	cgatsfile=new_cgats();
	if(type)
		cgatsfile->add_other(cgatsfile,type);
	if (cgatsfile->read_name(cgatsfile,filename)<0)
		throw cgatsfile->err;
}


CGATSFile::~CGATSFile()
{
	if(cgatsfile)
		cgatsfile->del(cgatsfile);
}


void CGATSFile::Save(const char *filename)
{
	if(cgatsfile->write_name(cgatsfile,filename)<0)
		throw cgatsfile->err;
}


const char *CGATSFile::GetKeyword(int table,const char *key)
{
	int i=cgatsfile->find_kword(cgatsfile,table,key);
	if(i>=0)
	{
		return(cgatsfile->t[table].kdata[i]);
	}
	return(NULL);
}


const char *CGATSFile::GetFieldName(int table,int i)
{
	return(cgatsfile->t[table].fsym[i]);
}


int CGATSFile::FindField(int table,const char *field)
{
	return(cgatsfile->find_field(cgatsfile,table,field));
}


////////////////////////////////////////////////////////////////////////////////////////////////////

// TIXFile handles the small amount of data we need from a TI1 file - namely the colorants and
// the order in which they appear.

TIXFile::TIXFile(const char *filename,const char *type)
	: CGATSFile(filename,type), DeviceNColorantList(), colorantstring(NULL), colorantoffset(0)
{
	const char *cols=GetKeyword(0,"COLOR_REP");
	colorantstring=strdup(cols);
	cerr << "COLOR_REP: " << colorantstring << endl;

	char *cn=colorantstring;
	while(*cn)
	{
		const char *longname=ArgyllSupport::GetColorantLongName(cn);
		if(longname)
		{
			cerr << "Found longname: " << longname << endl;
			new DeviceNColorant(*this,longname);
		}
		if(cn[0]=='2' || cn[0]=='3')
			++cn;

		if(cn[0]=='_')	// Break out of the loop if we reach '_'
			cn[0]=0;
		else
			++cn;
	}

	table=cgatsfile->t;
	sets=table->nsets;

	int nfields=table->nfields-1;
	for(int i=0;i<nfields;++i)
	{
		const char *fn=GetFieldName(i);
		if(strncmp(fn,colorantstring,strlen(colorantstring))==0)
		{
			colorantoffset=i;
			cerr << "Colorantoffset set to " << colorantoffset << endl;
			i=nfields;
		}
	}	
	for(int i=0;i<nfields;++i)
	{
		const char *fn=GetFieldName(i);
		if(strncmp(fn,"XYZ",3)==0)
		{
			colorants=i;
			i=nfields;
		}
	}
	Xindex=FindField("XYZ_X");
	Yindex=FindField("XYZ_Y");
	Zindex=FindField("XYZ_Z");

	cerr << colorants << " colorants and " << sets << " measurements" << endl;
}


TIXFile::~TIXFile()
{
	if(colorantstring)
		free(colorantstring);
}


static struct { const char *name; const char *alias;} fuzzymatches[]=
{
	{"Photo Black","Black"},
	{"Matte Black","Medium Black"},
	{NULL,NULL}
};


int TIXFile::GetColorantIndex(const char *name)
{
	if(!name)
		return(-1);
	int result=DeviceNColorantList::GetColorantIndex(name);
	if(result<0)
	{
		cerr << "Fuzzy matching " << name << endl;
		int idx=0;
		while(fuzzymatches[idx].name)
		{
			cerr << "...against " << fuzzymatches[idx].name << endl;
			if(StrcasecmpIgnoreSpaces(name,fuzzymatches[idx].name)==0)
			{
				result=DeviceNColorantList::GetColorantIndex(fuzzymatches[idx].alias);
				if(result>-1)
					cerr << name << " has index: " << result << endl;
				if(result>-1)
					return(result);
			}
			++idx;
		}
		throw "Colorant not found";
	}
	return(result);
}


int *TIXFile::CreateRawChannelMap(DeviceNColorantList &colorants)
{
	int colcount=colorants.GetColorantCount();
	int *channelmap=new int[colcount];
	for(int i=0;i<colcount;++i)
		channelmap[i]=0;

	DeviceNColorant *col=colorants.FirstColorant();
	int i=0;
	while(col)
	{
		try
		{
			int idx=GetColorantIndex(col->GetName());
			if(idx>-1)
				channelmap[idx]=i;
			else
				cerr << "-1: Channel not found" << endl;
		}
		catch(const char *err)
		{
			cerr << "Warning: " << err << endl;
		}
		col=col->NextColorant();
		++i;
	}
	for(i=0;i<colcount;++i)
		cerr << "Mapping " << i << " to " << channelmap[i] << endl;
	return(channelmap);
}


////////////////////////////////////////////////////////////////////////////////////////////////////

// TI3File handles specific aspects of Argyll's TI3 format.  Notably, field names and indices are
// offset, to ignore the SAMPLE_ID field.


TI3File::TI3File(const char *filename)
	: TIXFile(filename,"CTI3")
{
}

TI3File::~TI3File()
{
}

int TIXFile::GetColorantCount()
{
	return(colorants);
}

int TIXFile::GetSets()
{
	return(sets);
}

const char *TIXFile::GetFieldName(int i)
{
	return(CGATSFile::GetFieldName(0,i+colorantoffset));
}

int TIXFile::FindField(const char *field)
{
	// Offset by colorantoffset, to skip SAMPLE_ID
	return(CGATSFile::FindField(0,field)-colorantoffset);
}

double **TIXFile::GetData(int set)
{
	return(((double **)table->fdata[set])+colorantoffset);
}

